Imports System.Text.RegularExpressions

Public Class Action

    Private _number As Integer = 0
    Private _commandName As String = ""
    Private _parameters() As String = Nothing

    Public Sub New(ByVal number As Integer, ByVal commandName As String, ByVal parameters() As String)

        _number = number
        _commandName = commandName
        _parameters = CType(parameters.Clone(), String())

    End Sub

    Public ReadOnly Property Number() As Integer
        Get
            Return _number
        End Get
    End Property

    Public ReadOnly Property CommandName() As String
        Get
            Return _commandName
        End Get
    End Property

    Public ReadOnly Property ParamCount() As Integer
        Get
            Return _parameters.Length
        End Get
    End Property

    Public ReadOnly Property ParamValue(ByVal index As Integer) As String
        Get
            Return _parameters(index)
        End Get
    End Property

    Public Shared Function Parse(ByVal commandLine As String) As Action

        commandLine.Trim()

        Dim number As Integer = 0
        Dim commandName As String = ""
        Dim parameters() As String = Nothing

        With Regex.Match(commandLine, "(\d+)\s(.+)")
            If .Success Then
                number = [Integer].Parse(.Groups(1).Value.Trim())
                commandLine = .Groups(2).Value.Trim()
            Else
                Throw New FormatException([String].Format("Syntax Error at [{0}]", commandLine))
            End If
        End With

        With Regex.Match(commandLine, "(\w+)\s(.+)")
            If .Success Then
                commandName = .Groups(1).Value.Trim()
                commandLine = .Groups(2).Value.Trim()
            Else
                commandName = commandLine
            End If
        End With

        If commandLine <> "" Then

            Dim stringList As New ArrayList
            With New System.Text.StringBuilder(commandLine)
                .Append(" "c)
                Dim i As Integer = 0
                Do While i < .Length
                    If .Chars(i) = """"c Then
                        For j As Integer = i + 1 To .Length - 1
                            If .Chars(j) = """"c Then
                                If .Chars(j + 1) <> """"c Then
                                    Dim str As String = .ToString(i, j - i + 1)
                                    stringList.Add(str.Substring(1, str.Length - 2).Replace("""""", """"))
                                    .Remove(i, j - i + 1)
                                    .Insert(i, [String].Format("@{0}", stringList.Count - 1))
                                    Exit For
                                Else
                                    j += 1
                                End If
                            End If
                        Next j
                    End If
                    i += 1
                Loop
                commandLine = .ToString()
            End With

            commandLine = Regex.Replace(commandLine, ",", " ")
            commandLine = Regex.Replace(commandLine, "\s\s", " ")
            commandLine = commandLine.Trim()

            parameters = commandLine.Split()
            For i As Integer = 0 To parameters.Length - 1
                parameters(i) = parameters(i).Trim()
                If parameters(i).StartsWith("@") Then
                    parameters(i) = CStr(stringList([Integer].Parse(parameters(i).Substring(1))))
                End If
            Next i
        Else
            parameters = New String() {}
        End If

        Return New Action(number, commandName, parameters)

    End Function

    Public Enum Interrupting
        Enter
        Leave
        Contact
        Hit
    End Enum
End Class

Public NotInheritable Class ActionRunner

    'zWX^
    Private Structure Register

        Private _actionIndex As Integer '߃|C^
        Private _counter As Integer 'ėpJE^
        Private _callback As MethodInvoker '荞ݏ̏IɌĂяo܂B

        Public Property ActionIndex() As Integer
            Get
                Return _actionIndex
            End Get
            Set(ByVal Value As Integer)
                _actionIndex = Value
            End Set
        End Property

        Public Property Counter() As Integer
            Get
                Return _counter
            End Get
            Set(ByVal Value As Integer)
                _counter = Value
            End Set
        End Property

        Public Property Callback() As MethodInvoker
            Get
                Return _callback
            End Get
            Set(ByVal Value As MethodInvoker)
                _callback = Value
            End Set
        End Property
    End Structure

    Private CurrentState As Register = New Register '݂̃WX^
    Private StateStack As Stack = New Stack 'ۑꂽWX^

    'Enter ... Hero AActor  bounds ֊SɓB
    'Leave ... Hero AActor  bounds 犮SɏoB
    'Contact ... Hero A Actor ɑ΂Ęb|B܂ׂ͒B
    'Hit ... Hero ȀʍsłȂ Actor ɂԂB

    Private HandlerOnEnter As Integer = -1 'Enter 荞݂ɑ΂nh
    Private HandlerOnLeave As Integer = -1 'Leave 荞݂ɑ΂nh
    Private HandlerOnContact As Integer = -1 'Contact 荞݂ɑ΂nh
    Private HandlerOnHit As Integer = -1 'Hit 荞݂ɑ΂nh

    Private theActor As Actor = Nothing
    Private actions() As Action = Nothing

    Public Sub New(ByVal theActor As Actor, ByVal actions() As Action)

        Me.theActor = theActor
        Me.actions = CType(actions.Clone(), Action())

    End Sub

    '݁A荞݂ł True Ԃ܂B
    Public Function IsInterrupting() As Boolean

        If StateStack.Count > 0 Then
            Return True
        Else
            Return False
        End If

    End Function

    '݂̃WX^ԂۑAw肳ꂽCfbNXɎsڂ܂B
    '荞݃[`ւ̈ړ̍ۂɎgp܂B
    Private Sub [Call](ByVal index As Integer, ByVal callback As MethodInvoker)

        StateStack.Push(CurrentState)
        With CurrentState
            .ActionIndex = index
            .Counter = 0
            .Callback = callback
        End With

    End Sub

    'ŌɕۑꂽWX^Ԃ𕜌܂B
    Private Sub [Return]()

        CurrentState = CType(StateStack.Pop(), Register)

    End Sub

    'whatOn Ŏw肳ꂽ荞݂̃nhw肳Ă΁Å荞݂𔭐܂B
    'ꍇ True Ԃ܂B
    Public Function Interrupt(ByVal whatOn As Action.Interrupting, ByVal callback As MethodInvoker) As Boolean

        Select Case whatOn
            Case Action.Interrupting.Enter
                If HandlerOnEnter >= 0 Then
                    Me.Call(HandlerOnEnter, callback)
                    Return True
                End If
            Case Action.Interrupting.Leave
                If HandlerOnLeave >= 0 Then
                    Me.Call(HandlerOnLeave, callback)
                    Return True
                End If
            Case Action.Interrupting.Contact
                If HandlerOnContact >= 0 Then
                    Me.Call(HandlerOnContact, callback)
                    Return True
                End If
            Case Action.Interrupting.Hit
                If HandlerOnHit >= 0 Then
                    Me.Call(HandlerOnHit, callback)
                    Return True
                End If
        End Select

        Return False

    End Function

    'lineNumber Ŏw肳ꂽsԍANṼCfbNXԂ܂B
    Private Function GetActionIndex(ByVal lineNumber As Integer) As Integer

        For i As Integer = 0 To actions.Length - 1
            If actions(i).Number = lineNumber Then
                Return i
            End If
        Next i

        Return -1

    End Function

    '݂̃ANVs܂B
    'ɎsׂANVȂꍇAFalse Ԃ܂B
    Public Function DoAction() As Boolean

        If CurrentState.ActionIndex = actions.Length Then
            Return False
        End If

        With actions(CurrentState.ActionIndex)
            Select Case .CommandName.ToUpper()
                Case "MSGBOX"
                    MessageBox.Show(Instances.Current.Form, .ParamValue(0), "", MessageBoxButtons.OK, MessageBoxIcon.Information)
                    CurrentState.ActionIndex += 1
                Case "ON"
                    Select Case .ParamValue(0).ToUpper()
                        Case "ENTER"
                            Select Case .ParamValue(1).ToUpper()
                                Case "GOSUB"
                                    HandlerOnEnter = GetActionIndex([Integer].Parse(.ParamValue(2)))
                            End Select
                        Case "LEAVE"
                            Select Case .ParamValue(1).ToUpper()
                                Case "GOSUB"
                                    HandlerOnLeave = GetActionIndex([Integer].Parse(.ParamValue(2)))
                            End Select
                        Case "CONTACT"
                            Select Case .ParamValue(1).ToUpper()
                                Case "GOSUB"
                                    HandlerOnContact = GetActionIndex([Integer].Parse(.ParamValue(2)))
                            End Select
                        Case "HIT"
                            Select Case .ParamValue(1).ToUpper()
                                Case "GOSUB"
                                    HandlerOnHit = GetActionIndex([Integer].Parse(.ParamValue(2)))
                            End Select
                    End Select
                    CurrentState.ActionIndex += 1
                Case "GOTO"
                    CurrentState.ActionIndex = GetActionIndex([Integer].Parse(.ParamValue(0)))
                Case "RETURN"
                    If Not CurrentState.Callback Is Nothing Then
                        CurrentState.Callback.Invoke()
                    End If
                    Me.Return()
                Case "IF"
                    Select Case .ParamValue(0).ToUpper()
                        Case "DEFEATED"
                            If Instances.Current.GameInfo.IsDefeated(.ParamValue(1)) Then
                                Select Case .ParamValue(2).ToUpper()
                                    Case "THEN"
                                        CurrentState.ActionIndex = GetActionIndex([Integer].Parse(.ParamValue(3)))
                                    Case Else
                                        CurrentState.ActionIndex += 1
                                End Select
                            Else
                                If .ParamCount > 4 Then
                                    Select Case .ParamValue(4).ToUpper()
                                        Case "ELSE"
                                            CurrentState.ActionIndex = GetActionIndex([Integer].Parse(.ParamValue(5)))
                                        Case Else
                                            CurrentState.ActionIndex += 1
                                    End Select
                                Else
                                    CurrentState.ActionIndex += 1
                                End If
                            End If
                        Case "YESMSG"
                            If MessageBox.Show(Instances.Current.Form, .ParamValue(1), "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes Then
                                Select Case .ParamValue(2).ToUpper()
                                    Case "THEN"
                                        CurrentState.ActionIndex = GetActionIndex([Integer].Parse(.ParamValue(3)))
                                    Case Else
                                        CurrentState.ActionIndex += 1
                                End Select
                            Else
                                If .ParamCount > 4 Then
                                    Select Case .ParamValue(4).ToUpper()
                                        Case "ELSE"
                                            CurrentState.ActionIndex = GetActionIndex([Integer].Parse(.ParamValue(5)))
                                        Case Else
                                            CurrentState.ActionIndex += 1
                                    End Select
                                Else
                                    CurrentState.ActionIndex += 1
                                End If
                            End If
                        Case Else
                            CurrentState.ActionIndex += 1
                    End Select
                Case "STEPRIGHT"
                    theActor.Direction = Direction.Right
                    If CurrentState.Counter = 0 Then
                        Dim targetBounds As Rectangle = theActor.Bounds
                        targetBounds.Offset(32, 0)
                        If Instances.Current.CurrentArea.Contains(targetBounds) Then
                            Dim blockedActor As Actor = Nothing
                            If Instances.Current.CurrentArea.IsPassable(targetBounds, blockedActor) Then
                                theActor.Frames += 1
                                theActor.Move(8, 0)
                                CurrentState.Counter += 1
                                Exit Select
                            Else
                                If Not blockedactor Is Nothing Then
                                    blockedActor.Interrupt(Action.Interrupting.Hit, Nothing)
                                End If
                            End If
                        End If
                        CurrentState.ActionIndex += 1
                    ElseIf CurrentState.Counter < 3 Then
                        theActor.Frames += 1
                        theActor.Move(8, 0)
                        CurrentState.Counter += 1
                    Else
                        theActor.Frames = 0
                        theActor.Move(8, 0)
                        CurrentState.Counter = 0
                        CurrentState.ActionIndex += 1
                    End If
                Case "STEPLEFT"
                    theActor.Direction = Direction.Left
                    If CurrentState.Counter = 0 Then
                        Dim targetBounds As Rectangle = theActor.Bounds
                        targetBounds.Offset(-32, 0)
                        If Instances.Current.CurrentArea.Contains(targetBounds) Then
                            Dim blockedActor As Actor = Nothing
                            If Instances.Current.CurrentArea.IsPassable(targetBounds, blockedActor) Then
                                theActor.Frames += 1
                                theActor.Move(-8, 0)
                                CurrentState.Counter += 1
                                Exit Select
                            Else
                                If Not blockedactor Is Nothing Then
                                    blockedActor.Interrupt(Action.Interrupting.Hit, Nothing)
                                End If
                            End If
                        End If
                        CurrentState.ActionIndex += 1
                    ElseIf CurrentState.Counter < 3 Then
                        theActor.Frames += 1
                        theActor.Move(-8, 0)
                        CurrentState.Counter += 1
                    Else
                        theActor.Frames = 0
                        theActor.Move(-8, 0)
                        CurrentState.Counter = 0
                        CurrentState.ActionIndex += 1
                    End If
                Case "STEPDOWN"
                    theActor.Direction = Direction.Down
                    If CurrentState.Counter = 0 Then
                        Dim targetBounds As Rectangle = theActor.Bounds
                        targetBounds.Offset(0, 32)
                        If Instances.Current.CurrentArea.Contains(targetBounds) Then
                            Dim blockedActor As Actor = Nothing
                            If Instances.Current.CurrentArea.IsPassable(targetBounds, blockedActor) Then
                                theActor.Frames += 1
                                theActor.Move(0, 8)
                                CurrentState.Counter += 1
                                Exit Select
                            Else
                                If Not blockedactor Is Nothing Then
                                    blockedActor.Interrupt(Action.Interrupting.Hit, Nothing)
                                End If
                            End If
                        End If
                        CurrentState.ActionIndex += 1
                    ElseIf CurrentState.Counter < 3 Then
                        theActor.Frames += 1
                        theActor.Move(0, 8)
                        CurrentState.Counter += 1
                    Else
                        theActor.Frames = 0
                        theActor.Move(0, 8)
                        CurrentState.Counter = 0
                        CurrentState.ActionIndex += 1
                    End If
                Case "STEPUP"
                    theActor.Direction = Direction.Up
                    If CurrentState.Counter = 0 Then
                        Dim targetBounds As Rectangle = theActor.Bounds
                        targetBounds.Offset(0, -32)
                        If Instances.Current.CurrentArea.Contains(targetBounds) Then
                            Dim blockedActor As Actor = Nothing
                            If Instances.Current.CurrentArea.IsPassable(targetBounds, blockedActor) Then
                                theActor.Frames += 1
                                theActor.Move(0, -8)
                                CurrentState.Counter += 1
                                Exit Select
                            Else
                                If Not blockedactor Is Nothing Then
                                    blockedActor.Interrupt(Action.Interrupting.Hit, Nothing)
                                End If
                            End If
                        End If
                        CurrentState.ActionIndex += 1
                    ElseIf CurrentState.Counter < 3 Then
                        theActor.Frames += 1
                        theActor.Move(0, -8)
                        CurrentState.Counter += 1
                    Else
                        theActor.Frames = 0
                        theActor.Move(0, -8)
                        CurrentState.Counter = 0
                        CurrentState.ActionIndex += 1
                    End If
                Case "DOINN"
                    With Instances.Current
                        Dim fee As Integer = (.GameInfo.Win \ 100) * 100 + 100
                        If MessageBox.Show(.Form, [String].Format("͏hłBꔑ {0}$ łɂȂ܂H", fee), "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes Then
                            If .GameInfo.Money >= fee Then
                                .GameInfo.Money -= fee
                                .Renderer.FadeOut()
                                With .Hero
                                    .LP = .MaxLP
                                    .TP = .MaxTP
                                End With
                                .Wait(1000)
                                .Renderer.FadeIn()
                            Else
                                MessageBox.Show(.Form, "܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Error)
                            End If
                        End If
                        If MessageBox.Show(.Form, "f[^Z[u܂H", "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes Then
                            If .SaveData() = True Then
                                MessageBox.Show(.Form, "Z[u܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                            Else
                                MessageBox.Show(.Form, "炩̗RŃZ[uł܂łB", "", MessageBoxButtons.OK, MessageBoxIcon.Error)
                            End If
                        End If
                    End With
                    CurrentState.ActionIndex += 1
                Case "DOSHOP"
                    Dim tagList As New ArrayList
                    For i As Integer = 1 To .ParamCount - 1
                        tagList.Add(.ParamValue(i))
                    Next i
                    Instances.Current.Renderer.FadeOut()
                    With New Shop(.ParamValue(0), CType(tagList.ToArray(GetType(String)), String()))
                        .ShowDialog(Instances.Current.Form)
                        .Dispose()
                    End With
                    Instances.Current.Renderer.FadeIn()
                    CurrentState.ActionIndex += 1
                Case "DOFORGE"
                    If MessageBox.Show(Instances.Current.Form, "͒b艮łBb܂H", "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes Then
                        Dim theWeapon As Weapon = Instances.Current.Hero.Weapon
                        If Not theWeapon Is Nothing Then
                            Dim canForge As Boolean = False
                            For i As Integer = 0 To .ParamCount - 1
                                If theWeapon.Tag = .ParamValue(i) Then
                                    canForge = True
                                    Exit For
                                End If
                            Next i
                            If canForge Then
                                If theWeapon.OP < theWeapon.MaxOP Then
                                    Dim forge As Boolean = False
                                    Dim repair As Boolean = False
                                    Dim confirm As String = ""
                                    Dim fee As Integer = 0
                                    If theWeapon.Forged < theWeapon.MaxOP - theWeapon.BaseOP Then
                                        If theWeapon.Broken > 0 Then
                                            forge = True
                                            repair = True
                                            confirm = "Cp܂߂{0}$܂낵łH"
                                            fee = theWeapon.Price \ 2 + theWeapon.Broken * theWeapon.Price \ 10
                                        Else
                                            forge = True
                                            repair = False
                                            confirm = "̕b̂{0}$܂낵łH"
                                            fee = theWeapon.Price \ 2
                                        End If
                                    Else
                                        forge = False
                                        repair = True
                                        confirm = "̏̕Cɂ{0}$܂낵łH"
                                        fee = theWeapon.Broken * theWeapon.Price \ 10
                                    End If
                                    If MessageBox.Show(Instances.Current.Form, [String].Format(confirm, fee), "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes Then
                                        If Instances.Current.GameInfo.Money >= fee Then
                                            theWeapon.Broken = 0
                                            If forge Then
                                                theWeapon.Forged += 10
                                            End If
                                            Instances.Current.GameInfo.Money -= fee
                                            MessageBox.Show(Instances.Current.Form, "Ƃ܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                                        Else
                                            MessageBox.Show(Instances.Current.Form, "܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Error)
                                        End If
                                    End If
                                Else
                                    MessageBox.Show(Instances.Current.Form, "͂̕ȏ㋭ł܂񂵁ÂƂC̕Kv܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                                End If
                            Else
                                MessageBox.Show(Instances.Current.Form, "݂܂񂪁AM͖̕m̋ōĂ悤łB", "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                            End If
                        Else
                            MessageBox.Show(Instances.Current.Form, "𑕔Ă܂B", "", MessageBoxButtons.OK, MessageBoxIcon.Error)
                        End If
                    End If
                    CurrentState.ActionIndex += 1
                Case "DOWNDUNGEON"
                    Dim basement As Integer = [Integer].Parse(.ParamValue(0))
                    With Instances.Current
                        With .GameInfo
                            .Breakthrough = Math.Max(.Breakthrough, basement - 1)
                        End With
                        .ChangeArea(basement, True)
                    End With
                    CurrentState.ActionIndex += 1
                Case "UPDUNGEON"
                    Instances.Current.ChangeArea([Integer].Parse(.ParamValue(0)), False)
                    CurrentState.ActionIndex += 1
                Case "MOVEAREA"
                    Dim areaName As String = .ParamValue(0)
                    Dim locationX As Integer = [Integer].Parse(.ParamValue(1))
                    Dim locationY As Integer = [Integer].Parse(.ParamValue(2))
                    Instances.Current.ChangeArea(areaName, New Point(locationX, locationY))
                    CurrentState.ActionIndex += 1
                Case "FIGHTBOSS"
                    Dim bossName As String = .ParamValue(0)
                    If Instances.Current.DoFight(bossName) = True Then
                        Instances.Current.GameInfo.AddDefeated(bossName)
                    End If
                    CurrentState.ActionIndex += 1
                Case "HEALALL"
                    With Instances.Current
                        .Renderer.FlashScreen()
                        With .Hero
                            .LP = .MaxLP
                            .TP = .MaxTP
                        End With
                        MessageBox.Show(.Form, [String].Concat(.Hero.Name, "̂koƂsoAsȗ͂ɂĊSɉ񕜂I"), "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                    End With
                    CurrentState.ActionIndex += 1
                Case "GOENDING"
                    With Instances.Current
                        .Audio.Stop()
                        .Renderer.RenderEnding()
                        .ExitGame()
                    End With
                    CurrentState.ActionIndex += 1
            End Select
        End With

        If CurrentState.ActionIndex < actions.Length Then
            Return True
        Else
            Return False
        End If

    End Function
End Class
