Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions

Public Enum Direction
    None = 0
    Right = 1
    Left = 2
    Down = 3
    Up = 4
End Enum

Public NotInheritable Class ActorMovedEventArgs
    Inherits EventArgs

    Private _previousBounds As Rectangle = Rectangle.Empty
    Private _currentBounds As Rectangle = Rectangle.Empty

    Private Sub New()

    End Sub

    Public Sub New(ByVal previousBounds As Rectangle, ByVal currentBounds As Rectangle)

        _previousBounds = previousBounds
        _currentBounds = currentBounds

    End Sub

    Public ReadOnly Property PreviousBounds() As Rectangle
        Get
            Return _previousBounds
        End Get
    End Property

    Public ReadOnly Property CurrentBounds() As Rectangle
        Get
            Return _currentBounds
        End Get
    End Property
End Class

Public Class Actor

    Private _tag As String = "" '^O
    Private _skin As Integer = 0 'OtBbNX̔ԍ
    Private _bounds As Rectangle = Rectangle.Empty 'ꏊ
    Private _direction As Direction = Direction.Down '
    Private _frames As Integer = 0 '`pJE^

    Private runner As ActionRunner = New ActionRunner(Me, New Action() {}) 'ʏ펞̃ANV
    Private instantRunner As Queue = New Queue 'D悳ANV

    Public Event Moved(ByVal sender As Object, ByVal e As ActorMovedEventArgs)
    Protected Overridable Sub OnMoved(ByVal e As ActorMovedEventArgs)
        Try
            RaiseEvent Moved(Me, e)
        Finally

        End Try
    End Sub

    Public ReadOnly Property Tag() As String
        Get
            Return _tag
        End Get
    End Property

    Public Property Skin() As Integer
        Get
            Return _skin
        End Get
        Set(ByVal Value As Integer)
            _skin = Value
        End Set
    End Property

    Public Overridable Function GetSkinBounds() As Rectangle
        Return New Rectangle((_skin Mod 16) * 32, (_skin \ 16) * 32, 32, 32)
    End Function

    Public Property Bounds() As Rectangle
        Get
            Return _bounds
        End Get
        Set(ByVal Value As Rectangle)
            _bounds = Value
        End Set
    End Property

    Public Sub Move(ByVal x As Integer, ByVal y As Integer)
        Move(New Point(x, y))
    End Sub

    Public Sub Move(ByVal pos As Point)

        If Not pos.IsEmpty Then
            Dim previous As Rectangle = _bounds
            _bounds.Offset(pos)
            OnMoved(New ActorMovedEventArgs(previous, _bounds))
        End If

    End Sub

    Public Property Direction() As Direction
        Get
            Return _direction
        End Get
        Set(ByVal Value As Direction)
            _direction = Value
        End Set
    End Property

    Public Property Frames() As Integer
        Get
            Return _frames
        End Get
        Set(ByVal Value As Integer)
            _frames = Value
        End Set
    End Property

    Public Function IsWalking() As Boolean
        If _frames <> 0 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Function IsInterrupting() As Boolean
        Return runner.IsInterrupting()
    End Function

    Public Function Interrupt(ByVal whatOn As Action.Interrupting, ByVal callback As MethodInvoker) As Boolean
        Return runner.Interrupt(whatOn, callback)
    End Function

    Public Sub Proceed()

        If instantRunner.Count > 0 Then
            If CType(instantRunner.Peek(), ActionRunner).DoAction() = False Then
                instantRunner.Dequeue()
            End If
        Else
            runner.DoAction()
        End If

    End Sub

    Public Sub AddInstantAction(ByVal commandLine As String)
        instantRunner.Enqueue(New ActionRunner(Me, New Action() {Action.Parse(commandLine)}))
    End Sub

    Public Sub ClearInstantActions()
        instantRunner.Clear()
    End Sub

    Public Function HasInstantAction() As Boolean
        If instantRunner.Count > 0 Then
            Return True
        Else
            Return False
        End If
    End Function

    Public Sub ParseScript(ByVal filePath As String)

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            ParseScript(theFile)
        Finally
            theFile.Close()
        End Try

    End Sub

    Public Sub ParseScript(ByVal theStream As Stream)

        Dim commandLineList As New ArrayList

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                commandLineList.Add(Action.Parse(line))
                line = reader.ReadLine()
            Loop
        Finally
            reader.Close()
        End Try

        runner = New ActionRunner(Me, CType(commandLineList.ToArray(GetType(Action)), Action()))

    End Sub

    Protected Sub New(ByVal tagName As String)

        _tag = tagName

    End Sub

    Public Shared Function LoadData(ByVal filePath As String) As Object

        Dim actors() As Actor = Nothing

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            actors = CType(LoadData(theFile), Actor())
        Finally
            theFile.Close()
        End Try

        Dim actorDir As String = Path.GetDirectoryName(filePath)
        For Each anActor As Actor In actors
            anActor.ParseScript(Path.Combine(actorDir, [String].Concat(anActor.Tag, ".ini")))
        Next anActor

        Return actors

    End Function

    Private Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim actorList As New ArrayList

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = Regex.Replace(line, "\s", "")
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "(\w+),(\d+),\((\d+),(\d+)\)-\((\d+),(\d+)\)")
                        If .Success Then
                            Dim skin As Integer = [Integer].Parse(.Groups(2).Value)
                            Dim x1 As Integer = [Integer].Parse(.Groups(3).Value.Trim())
                            Dim y1 As Integer = [Integer].Parse(.Groups(4).Value.Trim())
                            Dim x2 As Integer = [Integer].Parse(.Groups(5).Value.Trim())
                            Dim y2 As Integer = [Integer].Parse(.Groups(6).Value.Trim())
                            Dim bounds As Rectangle = Area.AreaCoordToWorldCoord(Rectangle.FromLTRB( _
                                Min(x1, x2), Min(y1, y2), Max(x1, x2), Max(y1, y2)))
                            Dim newActor As New Actor(.Groups(1).Value)
                            With newActor
                                ._skin = skin
                                ._bounds = bounds
                            End With
                            actorList.Add(newActor)
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(actorList.ToArray(GetType(Actor)), Actor())

        Finally
            reader.Close()
        End Try

    End Function
End Class
