Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions
Imports GameLib

Public Enum SoundEffect
    Break = 0
    Hit = 1
    Shot = 2
End Enum

Public Enum BackgroundMusic
    Town
End Enum

Public Class AudioPlayer
    Implements IDisposable

    Private disposed As Boolean = False
    Private sounds() As OneShotSound = Nothing
    Private bgm As MIDIDevice = Nothing
    Private audioDev As AudioDevice = Nothing
    Private bgmData As Specialized.StringDictionary = New Specialized.StringDictionary

    Private Sub Dispose(ByVal disposing As Boolean)

        If Not disposed Then

            If disposing Then

                If Not sounds Is Nothing Then
                    For Each aSound As OneShotSound In sounds
                        If Not aSound Is Nothing Then
                            aSound.Dispose()
                        End If
                    Next aSound
                    sounds = Nothing
                End If

                If Not bgm Is Nothing Then
                    bgm.Dispose()
                    bgm = Nothing
                End If

                If Not audioDev Is Nothing Then
                    audioDev.Dispose()
                    audioDev = Nothing
                End If

            End If

        End If

        disposed = True

    End Sub

    Public Sub Dispose() Implements System.IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Public Sub New()

        Try
            audioDev = New AudioDevice
        Catch ex As Exception

        End Try

        If Not audioDev Is Nothing Then

            Dim soundNames() As String = [Enum].GetNames(GetType(SoundEffect))

            sounds = New OneShotSound(soundNames.Length - 1) {}
            For i As Integer = 0 To sounds.Length - 1
                sounds(i) = audioDev.CreateOneShotSound([String].Concat(soundNames(i), ".wav"))
            Next i

        End If

        Try
            bgm = New MIDIDevice
            bgm.SetupPort(MIDIPortFeatures.Shareable)
        Catch ex As Exception
            If Not bgm Is Nothing Then
                bgm.Dispose()
                bgm = Nothing
            End If
        End Try

    End Sub

    Public Sub Play(ByVal effect As SoundEffect)

        If Not audioDev Is Nothing Then
            audioDev.Play(sounds(effect))
        End If

    End Sub

    Public Const BGMNameInTown As String = "Town"
    Public Const BGMNameInDungeon As String = "Dungeon"
    Public Const BGMNameInNormalFight As String = "NormalFight"
    Public Const BGMNameInNormalBoss As String = "NormalBoss"
    Public Const BGMNameInFinalBoss As String = "FinalBoss"
    Public Const BGMNameInEnding As String = "Ending"

    Private _currentMusicName As String = ""

    Public ReadOnly Property CurrentMusicName() As String
        Get
            Return _currentMusicName
        End Get
    End Property

    Public Sub Play(ByVal musicName As String)

        If Not bgm Is Nothing Then
            If Not musicName Is Nothing AndAlso musicName <> "" Then
                If _currentMusicName <> musicName Then
                    Dim data() As String = bgmData(musicName).Split(","c)
                    Dim fileName As String = [String].Concat(data(0).Trim())
                    Dim loopStart As Integer = [Integer].Parse(data(1).Trim())
                    Dim loopEnd As Integer = [Integer].Parse(data(2).Trim())

                    bgm.Stop()
                    bgm.Play(fileName, loopStart, loopEnd)
                    _currentMusicName = musicName
                End If
            Else
                Me.Stop()
            End If
        End If

    End Sub

    Public Sub [Stop]()

        If Not bgm Is Nothing Then
            bgm.Stop()
        End If

        _currentMusicName = ""

    End Sub

    Public Function IsPlayingBGM() As Boolean
        If Not bgm Is Nothing Then
            Return bgm.IsPlaying()
        Else
            Return False
        End If
    End Function

    Public Sub LoadBGMConfig(ByVal filePath As String)

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            LoadBGMConfig(theFile)
        Finally
            theFile.Close()
        End Try

    End Sub

    Public Sub LoadBGMConfig(ByVal theStream As Stream)

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "(.+)=(.+)")
                        If .Success Then
                            bgmData.Add(.Groups(1).Value.Trim(), .Groups(2).Value)
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

        Finally
            reader.Close()
        End Try
    End Sub

End Class
