Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions

Public Interface IFighter

    ReadOnly Property MaxLP() As Integer
    ReadOnly Property MaxTP() As Integer
    Property LP() As Integer
    Property TP() As Integer
    ReadOnly Property OP() As Integer
    ReadOnly Property DP() As Integer
    ReadOnly Property MV() As Integer

End Interface

Public Class Enemy
    Implements IFighter

    Private _tag As String = ""
    Private kind As Integer = 0

    Private _maxLP As Integer = 0
    Private _lifePoint As Integer = 0
    Private _op As Integer = 0
    Private _dp As Integer = 0
    Private _mv As Integer = 0
    Private _money As Integer = 0

    Protected Overridable Sub CopyTo(ByVal dest As Enemy)

        With dest
            ._tag = _tag
            .kind = kind
            ._maxLP = _maxLP
            ._lifePoint = _lifePoint
            ._op = _op
            ._dp = _dp
            ._mv = _mv
            ._money = _money
        End With

    End Sub

    Public ReadOnly Property Tag() As String
        Get
            Return _tag
        End Get
    End Property

    Public Overridable Function GetSkinBounds() As Rectangle
        Return New Rectangle(64 + 64 * (kind Mod 4), 256 + 64 * (kind \ 4), 64, 64)
    End Function

    Public ReadOnly Property MaxLP() As Integer Implements IFighter.MaxLP
        Get
            Return _maxLP
        End Get
    End Property

    Protected Sub SetMaxLP(ByVal value As Integer)
        _maxLP = Max(0, value)
    End Sub

    Public ReadOnly Property MaxTP() As Integer Implements IFighter.MaxTP
        Get
            Return 0
        End Get
    End Property

    Public Property LP() As Integer Implements IFighter.LP
        Get
            Return _lifePoint
        End Get
        Set(ByVal Value As Integer)
            _lifePoint = Min(Max(0, Value), MaxLP)
        End Set
    End Property

    Public Property TP() As Integer Implements IFighter.TP
        Get
            Return 0
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property

    Public ReadOnly Property OP() As Integer Implements IFighter.OP
        Get
            Return _op
        End Get
    End Property

    Protected Sub SetOP(ByVal value As Integer)
        _op = value
    End Sub

    Public ReadOnly Property DP() As Integer Implements IFighter.DP
        Get
            Return _dp
        End Get
    End Property

    Protected Sub SetDP(ByVal value As Integer)
        _dp = value
    End Sub

    Public ReadOnly Property MV() As Integer Implements IFighter.MV
        Get
            Return _mv
        End Get
    End Property

    Protected Sub SetMV(ByVal value As Integer)
        _mv = value
    End Sub

    Public ReadOnly Property Money() As Integer
        Get
            Return _money
        End Get
    End Property

    Protected Sub SetMoney(ByVal value As Integer)
        _money = value
    End Sub

    Protected Sub New(ByVal tagName As String)

        _tag = tagName

    End Sub

    Public Sub New(ByVal kind As Integer)

        Me.kind = kind

        Dim a As Integer = 0
        With Instances.Current.Hero
            a = (.PhysicalEnergy - .PhysicalEnergyBase) + (.PsychicEnergy - .PsychicEnergyBase) + _
                (.Strength - .StrengthBase) + (.Quickness - .QuicknessBase)
        End With

        _maxLP = 50 + a
        _lifePoint = _maxLP
        _op = 50 + a * 3 \ 4
        _dp = 30 + a * 3 \ 4
        _mv = 20 + a * 3 \ 4
        _money = 100 + Instances.Current.GameInfo.Fight

        Select Case kind
            Case 0

            Case 1
                _op = _op * 13 \ 10
            Case 2
                _dp = _dp * 13 \ 10
            Case 3
                _mv = _mv * 13 \ 10
            Case 4
                _op = _op * 12 \ 10
                _dp = _dp * 12 \ 10
            Case 5
                _dp = _dp * 12 \ 10
                _mv = _mv * 12 \ 10
            Case 6
                _mv = _mv * 12 \ 10
                _op = _op * 12 \ 10
            Case 7
                _op = _op * 11 \ 10
                _dp = _dp * 11 \ 10
                _mv = _mv * 11 \ 10
            Case Else
                Throw New ArgumentOutOfRangeException("Enemy kind must be within 0 to 7.")
        End Select

    End Sub
End Class

Public Class Boss
    Inherits Enemy

    Private skinBounds As Rectangle = New Rectangle(0, 0, 128, 128)
    Private _bgmName As String = ""
    Private beginMsgList As ArrayList = New ArrayList
    Private finishMsgList As ArrayList = New ArrayList

    Protected Overrides Sub CopyTo(ByVal dest As Enemy)
        MyBase.CopyTo(dest)

        With CType(dest, Boss)
            .skinBounds = skinBounds
            .beginMsgList = CType(beginMsgList.Clone(), ArrayList)
            .finishMsgList = CType(finishMsgList.Clone(), ArrayList)
        End With

    End Sub

    Public Function Clone() As Boss

        Dim dest As New Boss(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Overrides Function GetSkinBounds() As Rectangle
        Return skinBounds
    End Function

    Public Property BGMName() As String
        Get
            Return _bgmName
        End Get
        Set(ByVal Value As String)
            _bgmName = Value
        End Set
    End Property

    Public Function GetBeginMsgs() As String()
        Return CType(beginMsgList.ToArray(GetType(String)), String())
    End Function

    Public Function GetFinishMsgs() As String()
        Return CType(finishMsgList.ToArray(GetType(String)), String())
    End Function

    Protected Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim bossList As New ArrayList
        Dim newBoss As Boss = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newBoss = New Boss(.Groups(1).Value.Trim())
                            bossList.Add(newBoss)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newBoss
                                        Select Case paramName.ToUpper()
                                            Case "LP"
                                                .SetMaxLP([Integer].Parse(paramValue))
                                                .LP = .MaxLP
                                            Case "OP"
                                                .SetOP([Integer].Parse(paramValue))
                                            Case "DP"
                                                .SetDP([Integer].Parse(paramValue))
                                            Case "MV"
                                                .SetMV([Integer].Parse(paramValue))
                                            Case "MONEY"
                                                .SetMoney([Integer].Parse(paramValue))
                                            Case "SKINLEFT"
                                                .skinBounds.X = [Integer].Parse(paramValue)
                                            Case "SKINTOP"
                                                .skinBounds.Y = [Integer].Parse(paramValue)
                                            Case "BGM"
                                                ._bgmName = paramValue
                                            Case "BEGINMSG"
                                                .beginMsgList.Add(paramValue)
                                            Case "FINISHMSG"
                                                .finishMsgList.Add(paramValue)
                                        End Select
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(bossList.ToArray(GetType(Boss)), Boss())

        Finally
            reader.Close()
        End Try

    End Function
End Class
