Public Class FightManager
    Implements IAdditionalRenderer

    Private currentEnemy As Enemy = Nothing

    Private _fightInfo As String = ""
    Private _bigMsg As String = ""

    Public ReadOnly Property Enemy() As Enemy
        Get
            Return currentEnemy
        End Get
    End Property

    Public Property FightInfo() As String
        Get
            Return _fightInfo
        End Get
        Set(ByVal Value As String)
            _fightInfo = Value
        End Set
    End Property

    Public Property BigMsg() As String
        Get
            Return _bigMsg
        End Get
        Set(ByVal Value As String)
            _bigMsg = Value
        End Set
    End Property

    Public Sub New(ByVal theEnemy As Enemy)

        currentEnemy = theEnemy

    End Sub

    Private Enum HeroCommand
        Normal
        Tech
        Escape
        Tool
    End Enum

    Private MustInherit Class ListBox

        Private items() As String = {}
        Private index As Integer = 0
        Private visible As Boolean = False
        Private decided As Boolean = False
        Private canceled As Boolean = False

        Protected Sub SetItems(ByVal items() As String)
            Me.items = items
        End Sub

        Public ReadOnly Property SelectedIndex() As Integer
            Get
                Return index
            End Get
        End Property

        Public Property IsVisible() As Boolean
            Get
                Return visible
            End Get
            Set(ByVal Value As Boolean)
                visible = Value
            End Set
        End Property

        Public Overridable Sub Reset()

            index = 0
            decided = False
            canceled = False

        End Sub

        Protected Sub Decide()
            decided = True
        End Sub

        Public Property IsCanceled() As Boolean
            Get
                Return canceled
            End Get
            Set(ByVal Value As Boolean)
                canceled = Value
            End Set
        End Property

        Protected MustOverride Sub OnSelected()
        Protected MustOverride Sub OnBack()

        Private Sub KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)

            If e.KeyCode = Keys.Down Then
                If index < items.Length - 1 Then
                    index += 1
                End If
            ElseIf e.KeyCode = Keys.Up Then
                If 0 < index Then
                    index -= 1
                End If
            ElseIf e.KeyCode = Keys.Space Then
                OnSelected()
            ElseIf e.KeyCode = Keys.Back Then
                OnBack()
            End If

        End Sub

        Public MustOverride Sub Render(ByVal g As Graphics)

        Protected Sub GetInputFocus()

            AddHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

        End Sub

        Protected Sub ReleaseInputFocus()

            RemoveHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

        End Sub

        Public Sub ShowDialog()

            GetInputFocus()
            visible = True
            Do Until decided
                Instances.Current.Wait(50)
                Application.DoEvents()
                Instances.Current.Renderer.RefreshScreen()
            Loop
            visible = False
            ReleaseInputFocus()

        End Sub
    End Class

    Private Class CommandListBox
        Inherits ListBox

        Private listItems() As String = {"", "Z", "", "ACe"}

        Public Sub New()

            SetItems(listItems)

        End Sub

        Private Class ItemListBox
            Inherits ListBox

            Private listItems() As String = {}

            Private Sub RefreshList()

                listItems = New String(Instances.Current.Hero.Tools.Count - 1) {}
                For i As Integer = 0 To listItems.Length - 1
                    listItems(i) = Instances.Current.Hero.Tools(i).Name
                Next i
                SetItems(listItems)

            End Sub

            Public Sub New()

                RefreshList()

            End Sub

            Public Overrides Sub Reset()
                MyBase.Reset()

                RefreshList()

            End Sub

            Protected Overrides Sub OnSelected()

                IsCanceled = False
                Decide()

            End Sub

            Protected Overrides Sub OnBack()

                IsCanceled = True
                Decide()

            End Sub

            Public Overrides Sub Render(ByVal g As System.Drawing.Graphics)

                If IsVisible Then
                    Dim font As New Font("lr oSVbN", 12.0F)
                    g.DrawRectangle(Pens.Red, New Rectangle(98, 72, 82, 20 * listItems.Length + 2))
                    g.FillRectangle(Brushes.DarkRed, New Rectangle(99, 73 + 20 * SelectedIndex, 80, 20))
                    For i As Integer = 0 To listItems.Length - 1
                        g.DrawString(listItems(i), font, Brushes.White, 99, 73 + 20 * i)
                    Next i
                    font.Dispose()
                End If

            End Sub
        End Class

        Private itemList As ItemListBox = New ItemListBox

        Public ReadOnly Property ItemListIndex() As Integer
            Get
                Return itemList.SelectedIndex
            End Get
        End Property

        Protected Overrides Sub OnSelected()

            If SelectedIndex <= 2 Then
                Decide()
            Else
                If Instances.Current.Hero.Tools.Count > 0 Then
                    With itemList
                        .Reset()
                        .IsVisible = True
                        Me.ReleaseInputFocus()
                        .ShowDialog()
                        Me.GetInputFocus()
                        .IsVisible = False
                        If Not .IsCanceled Then
                            Decide()
                        End If
                    End With
                Else
                    Instances.Current.Fight.FightInfo = "ACe͂ЂƂ܂"
                End If
            End If

        End Sub

        Protected Overrides Sub OnBack()

        End Sub

        Public Overrides Sub Render(ByVal g As Graphics)

            If IsVisible Then
                If Not itemList.IsVisible Then
                    g.DrawRectangle(Pens.Red, New Rectangle(16, 72, 82, 82))
                    g.FillRectangle(Brushes.DarkRed, New Rectangle(17, 73 + 20 * SelectedIndex, 80, 20))

                    Dim font As New Font("lr oSVbN", 12.0F)
                    For i As Integer = 0 To listItems.Length - 1
                        g.DrawString(CStr(listItems(i)), font, Brushes.White, 17, 73 + 20 * i)
                    Next i
                    font.Dispose()
                Else
                    g.DrawRectangle(Pens.Red, New Rectangle(16, 72, 82, 82))
                    g.FillRectangle(Brushes.Gray, New Rectangle(17, 73 + 20 * SelectedIndex, 80, 20))

                    Dim font As New Font("lr oSVbN", 12.0F)
                    For i As Integer = 0 To listItems.Length - 1
                        g.DrawString(CStr(listItems(i)), font, Brushes.DarkGray, 17, 73 + 20 * i)
                    Next i
                    font.Dispose()
                End If
            End If

            itemList.Render(g)

        End Sub

    End Class

    Private commandList As CommandListBox = New CommandListBox
    Private smallFont As Font = Nothing
    Private mediumFont As Font = Nothing
    Private largeFont As Font = Nothing

    Public Sub Render(ByVal g As Graphics) Implements IAdditionalRenderer.Render

        g.DrawString("LP", smallFont, Brushes.LightCoral, 8.0F, 7.0F)
        g.DrawString("TP", smallFont, Brushes.LightCoral, 8.0F, 30.0F)
        g.DrawString("OP", smallFont, Brushes.LightCoral, 120.0F, 15.0F)
        g.DrawString("DP", smallFont, Brushes.LightCoral, 152.0F, 15.0F)
        g.DrawString("MV", smallFont, Brushes.LightCoral, 184.0F, 15.0F)

        g.DrawString([String].Format( _
            "{0} / {1}", Instances.Current.Hero.LP, Instances.Current.Hero.MaxLP), _
            smallFont, Brushes.LightGreen, 32.0F, 7.0F)
        g.DrawString([String].Format( _
            "{0} / {1}", Instances.Current.Hero.TP, Instances.Current.Hero.MaxTP), _
            smallFont, Brushes.LightGreen, 32.0F, 30.0F)
        g.DrawString(Instances.Current.Hero.OP.ToString(), _
            smallFont, Brushes.LightGreen, 120.0F, 30.0F)
        g.DrawString(Instances.Current.Hero.DP.ToString(), _
            smallFont, Brushes.LightGreen, 152.0F, 30.0F)
        g.DrawString(Instances.Current.Hero.MV.ToString(), _
            smallFont, Brushes.LightGreen, 184.0F, 30.0F)

        g.DrawString(FightInfo, mediumFont, Brushes.LightGreen, 0.0F, 336.0F)

        g.DrawLine(Pens.LightCoral, 0, 52, 216, 52)
        g.DrawLine(Pens.LightCoral, 216, 0, 216, 52)
        g.DrawLine(Pens.LightCoral, 0, 332, 640, 332)

        commandList.Render(g)

        g.DrawString(BigMsg, Instances.Current.Fight.largeFont, Brushes.SkyBlue, 100.0F, 100.0F)

    End Sub

    Private Class NormalAttackManager
        Implements IAdditionalRenderer

        Private power As Integer = 0
        Private hit As Boolean = False

        Public Sub DoMotion()

            Instances.Current.Fight.BigMsg = "Stop Near 100!"

            power = 0
            hit = False

            AddHandler Instances.Current.Form.KeyDown, AddressOf KeyDown
            Instances.Current.Renderer.AddRenderer(Me)

            Dim powerDelta As Double = (Instances.Current.Fight.Enemy.MV / Instances.Current.Hero.MV) / 5
            Dim powerExact As Double = 0.0
            Do Until hit
                powerExact += powerDelta
                power = CInt(Fix(powerExact))
                If power > 100 Then
                    power = 0
                    Exit Do
                End If
                Instances.Current.Renderer.RefreshScreen()
                Instances.Current.Wait(10)
            Loop

            Instances.Current.Renderer.RemoveRenderer(Me)
            RemoveHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

            Instances.Current.Fight.BigMsg = ""

            Instances.Current.Renderer.RefreshScreen()

            If power = 100 Then
                Instances.Current.Renderer.DoAnimation(New String() {"FL 300"})
            End If

            Dim motions() As String = { _
                "HM 5", _
                "WT 100", _
                "HM 0", _
                "WT 100", _
                "HM 1", _
                "DE 3, 0, 0", _
                "SE 1", _
                "WT 500" _
            }

            Instances.Current.Renderer.DoAnimation(motions)

            With Instances.Current
                Dim damage As Integer = 0
                If power < 100 Then
                    damage = Math.Max(0, .Hero.OP * (200 + power) \ 200 - .Fight.Enemy.DP * 3 \ 4)
                Else
                    damage = .Hero.OP * (200 + power) \ 200
                End If
                damage = damage * (New Random).Next(95, 105) \ 100
                .Fight.Enemy.LP -= damage
                .Fight.FightInfo = [String].Format("G{0}̃_[WI", damage)
                If Not .Hero.Weapon Is Nothing Then
                    If (New Random).Next(0, 100) < 5 Then
                        .Hero.Weapon.Broken += 1
                    End If
                End If
                .Renderer.RefreshScreen()
                .Wait(1000)
            End With

        End Sub

        Private Sub KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)

            If e.KeyCode = Keys.Space Then
                hit = True
            End If

        End Sub

        Private Sub Render(ByVal g As Graphics) Implements IAdditionalRenderer.Render

            g.DrawRectangle(Pens.Gray, New Rectangle(208, 164, 204, 24))
            g.FillRectangle(Brushes.SkyBlue, New Rectangle(209, 165, power * 202 \ 100, 22))

        End Sub
    End Class

    Private Class TechManager

        Private index As Integer = 0
        Private spell As String = ""
        Private done As Boolean = False

        Public ReadOnly Property HasDone() As Boolean
            Get
                Return done
            End Get
        End Property

        Public Sub DoMotion()

            With Instances.Current
                If .Hero.TP = 0 Then
                    .Fight.FightInfo = "Zg󋵂ɂ܂B"
                    .Renderer.RefreshScreen()
                    .Wait(1000)
                    done = False
                    Return
                End If
            End With

            Instances.Current.Fight.BigMsg = "Input the Code!"

            Dim mulOP As Integer = 100
            Dim mulDP As Integer = 100
            Dim motions() As String = New String() {}

            With Instances.Current.Hero
                If Not .Weapon Is Nothing Then
                    With .Weapon
                        If Not .CanUseTech2 Then
                            With .Tech1
                                spell = .Name
                                mulOP = .MulOP
                                mulDP = .MulDP
                                motions = .GetMotions()
                            End With
                        Else
                            With .Tech2
                                spell = .Name
                                mulOP = .MulOP
                                mulDP = .MulDP
                                motions = .GetMotions()
                            End With
                        End If
                    End With
                Else
                    spell = "Bakuretuken"
                    mulOP = 200
                    mulDP = 100
                    motions = New String() { _
                        "FL 100", _
                        "DE 5, 0, 0", _
                        "SE 0", _
                        "WT 1000" _
                    }
                End If
            End With

            index = 0

            AddHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

            Dim succeeded As Boolean = False

            Dim powerDelta As Double = 0.1
            Dim exactTP As Double = Instances.Current.Hero.TP
            Do
                exactTP -= powerDelta
                Instances.Current.Hero.TP = CInt(Fix(exactTP))
                Instances.Current.Renderer.RefreshScreen()
                Instances.Current.Wait(10)
                If index = spell.Length Then
                    succeeded = True
                    Exit Do
                ElseIf Instances.Current.Hero.TP = 0 Then
                    succeeded = False
                    Exit Do
                End If
            Loop

            RemoveHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

            Instances.Current.Fight.BigMsg = ""

            Instances.Current.Renderer.RefreshScreen()

            If Not succeeded Then
                With Instances.Current
                    .Fight.FightInfo = "͂~XI"
                    Instances.Current.Renderer.RefreshScreen()
                    .Wait(1000)
                End With
                done = True
                Return
            End If

            Instances.Current.Renderer.DoAnimation(motions)

            With Instances.Current
                Dim damage As Integer = Math.Max(0, (.Hero.OP * mulOP \ 100) - (.Fight.Enemy.DP * mulDP \ 100) * 3 \ 4)
                damage = damage * (New Random).Next(95, 105) \ 100
                .Fight.Enemy.LP -= damage
                .Fight.FightInfo = [String].Format("G{0}̃_[WI", damage)
                If Not .Hero.Weapon Is Nothing Then
                    If (New Random).Next(0, 100) < 5 Then
                        .Hero.Weapon.Broken += 1
                    End If
                End If
                .Renderer.RefreshScreen()
                .Wait(1000)
            End With

            done = True

        End Sub

        Private Sub KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)

            If AscW([Char].ToUpper(spell.Chars(index))) = e.KeyCode Then
                Instances.Current.Audio.Play(SoundEffect.Shot)
                index += 1
                Instances.Current.Fight.BigMsg = spell.Substring(0, index)
            End If

        End Sub
    End Class

    Private Class EscapeManager
        Implements IAdditionalRenderer

        Private power As Integer = 0
        Private hit As Boolean = False
        Private _succeeded As Boolean = False

        Public ReadOnly Property Succeeded() As Boolean
            Get
                Return _succeeded
            End Get
        End Property

        Public Sub DoMotion()

            Instances.Current.Fight.BigMsg = "Stop Near 100!"

            power = 0
            hit = False

            AddHandler Instances.Current.Form.KeyDown, AddressOf KeyDown
            Instances.Current.Renderer.AddRenderer(Me)

            Dim powerDelta As Double = (Instances.Current.Fight.Enemy.MV / Instances.Current.Hero.MV) / 5
            Dim powerExact As Double = 0.0
            Do Until hit
                powerExact += powerDelta
                power = CInt(Fix(powerExact))
                If power > 100 Then
                    power = 0
                    Exit Do
                End If
                Instances.Current.Renderer.RefreshScreen()
                Instances.Current.Wait(10)
            Loop

            Instances.Current.Renderer.RemoveRenderer(Me)
            RemoveHandler Instances.Current.Form.KeyDown, AddressOf KeyDown

            Instances.Current.Fight.BigMsg = ""

            With Instances.Current
                If .Fight.Enemy.MV - .Hero.MV <= power Then
                    .Fight.FightInfo = [String].Concat(.Hero.Name, "͓oI")
                    .Audio.Stop()
                    .Renderer.RefreshScreen()
                    .Wait(1000)
                    _succeeded = True
                Else
                    .Fight.FightInfo = "GɑޘHӂꂽI"
                    .Renderer.RefreshScreen()
                    .Wait(1000)
                    _succeeded = False
                End If
            End With

        End Sub

        Private Sub KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)

            If e.KeyCode = Keys.Space Then
                hit = True
            End If

        End Sub

        Private Sub Render(ByVal g As Graphics) Implements IAdditionalRenderer.Render

            g.DrawRectangle(Pens.Gray, New Rectangle(208, 164, 204, 24))
            g.FillRectangle(Brushes.SkyBlue, New Rectangle(209, 165, power * 202 \ 100, 22))

        End Sub
    End Class

    Private Class ToolManager

        Private index As Integer = 0
        Private done As Boolean = False

        Public Sub New(ByVal index As Integer)
            Me.index = index
        End Sub

        Public ReadOnly Property HasDone() As Boolean
            Get
                Return done
            End Get
        End Property

        Public Sub DoMotion()

            Instances.Current.Fight.BigMsg = ""

            Dim theTool As Item = Instances.Current.Hero.Tools(index)

            If GetType(LPHealer).IsInstanceOfType(theTool) Then
                With Instances.Current
                    .Hero.LP += CType(theTool, LPHealer).Recovery
                    .Fight.FightInfo = "LPsRɉ񕜂I"
                    .Hero.Tools.RemoveAt(index)
                    done = True
                End With
            ElseIf GetType(Bomb).IsInstanceOfType(theTool) Then
                Dim bomb As Bomb = CType(theTool, Bomb)
                With Instances.Current
                    If (New Random).Next(0, 100) >= bomb.Accidentality Then
                        .Fight.Enemy.LP -= bomb.Damage * (New Random).Next(95, 105) \ 100

                        Dim motions() As String = { _
                            "FL 100", _
                            "DE 5, 0, 0", _
                            "SE 0", _
                            "WT 1000" _
                        }
                        Instances.Current.Renderer.DoAnimation(motions)

                        .Fight.FightInfo = [String].Format("G{0}̃_[W^ȂI", bomb.Damage)
                    Else
                        .Hero.LP -= bomb.Damage

                        Dim motions() As String = { _
                            "DH 5, 0, 0" _
                        }
                        Instances.Current.Renderer.DoAnimation(motions)

                        .Fight.FightInfo = "\I"
                    End If
                    .Hero.Tools.RemoveAt(index)
                    done = True
                End With
            Else
                Instances.Current.Fight.FightInfo = "gĂꍇł͂܂B"
                done = False
            End If

            Instances.Current.Renderer.RefreshScreen()
            Instances.Current.Wait(1000)

        End Sub
    End Class

    Private Class EnemyAttackManager

        Private inputIsValid As Boolean = True
        Private inputting As Boolean = False
        Private hit As Boolean = False

        Public Sub DoMotion()

            With Instances.Current
                .Fight.FightInfo = "G̍UI"
                .Renderer.RefreshScreen()
                AddHandler .Form.KeyDown, AddressOf KeyDown
                AddHandler .Form.KeyUp, AddressOf KeyUp
                .Wait(1000)
                inputting = True
                .Wait(50)
                inputting = False
                RemoveHandler .Form.KeyDown, AddressOf KeyDown
                RemoveHandler .Form.KeyUp, AddressOf KeyUp
            End With

            If Not hit Then
                If (New Random).Next(0, 10) > 0 Then
                    With Instances.Current
                        Dim motions() As String = { _
                            "HM 3", _
                            "DH 3, 0, 0", _
                            "SE 1", _
                            "WT 500" _
                        }
                        .Renderer.DoAnimation(motions)

                        Dim damage As Integer = Math.Max(0, .Fight.Enemy.OP - .Hero.DP * 3 \ 4)
                        damage = damage * (New Random).Next(95, 105) \ 100
                        .Hero.LP -= damage
                        .Fight.FightInfo = [String].Format("{0}̃_[W󂯂I", damage)
                        .Renderer.RefreshScreen()
                        .Wait(1000)
                    End With
                Else
                    With Instances.Current
                        Dim motions() As String = { _
                            "FL 300", _
                            "HM 4", _
                            "DH 1, 0, 0", _
                            "SE 1", _
                            "WT 500" _
                        }
                        .Renderer.DoAnimation(motions)

                        Dim damage As Integer = .Fight.Enemy.OP
                        damage = damage * (New Random).Next(95, 105) \ 100
                        .Hero.LP -= damage
                        .Fight.FightInfo = [String].Format("G̟Ӑg̈ꌂI {0}̃_[W󂯂I", damage)
                        .Renderer.RefreshScreen()
                        .Wait(1000)
                    End With
                End If
            Else
                With Instances.Current
                    Dim motions() As String = { _
                        "HM 3", _
                        "WT 300", _
                        "FL 300", _
                        "HM 5", _
                        "WT 100", _
                        "HM 0", _
                        "WT 100", _
                        "HM 1", _
                        "DE 3, 0, 0", _
                        "SE 1", _
                        "WT 500" _
                    }
                    .Renderer.DoAnimation(motions)

                    Dim damage As Integer = Math.Max(0, .Hero.OP - .Fight.Enemy.DP * 3 \ 4)
                    damage = damage * (New Random).Next(95, 105) \ 100
                    .Fight.Enemy.LP -= damage
                    .Fight.FightInfo = [String].Format("JE^[A^bNI@G{0}̃_[WI", damage)
                    .Renderer.RefreshScreen()
                    .Wait(1000)
                End With
            End If

        End Sub

        Private Sub KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)

            If e.KeyCode = Keys.Space Then
                If inputting Then
                    If inputIsValid Then
                        hit = True
                    End If
                Else
                    inputIsValid = False
                End If
            End If

        End Sub

        Private Sub KeyUp(ByVal sender As Object, ByVal e As KeyEventArgs)

            If e.KeyCode = Keys.Space Then
                inputIsValid = True
            End If

        End Sub
    End Class

    Private won As Boolean = False

    Public ReadOnly Property HeroWon() As Boolean
        Get
            Return won
        End Get
    End Property

    Public Sub Fire()

        With Instances.Current
            .Renderer.RenderedObjectInFight = SceneRenderer.VisibilityInFight.Everything
            .GameInfo.Fight += 1
        End With

        smallFont = New Font("MS UI Gothic", 9.0F)
        mediumFont = New Font("MS UI Gothic", 12.0F)
        largeFont = New Font("lr oSVbN", 24.0F)

        Instances.Current.Renderer.AddRenderer(Me)

        Dim heroTurn As Boolean = True

        Do
            If heroTurn Then

                BigMsg = "R}hÎI"
                FightInfo = ""

                commandList.Reset()
                commandList.ShowDialog()
                Select Case commandList.SelectedIndex
                    Case 0
                        With New NormalAttackManager
                            .DoMotion()
                        End With
                        heroTurn = False
                    Case 1
                        With New TechManager
                            .DoMotion()
                            If .HasDone Then
                                heroTurn = False
                            End If
                        End With
                    Case 2
                        With New EscapeManager
                            .DoMotion()
                            If .Succeeded Then
                                Exit Do
                            End If
                            heroTurn = False
                        End With
                    Case 3
                        With New ToolManager(commandList.ItemListIndex)
                            .DoMotion()
                            If .HasDone Then
                                heroTurn = False
                            End If
                        End With
                End Select
            Else
                With New EnemyAttackManager
                    .DoMotion()
                End With
                heroTurn = True
            End If

            If Instances.Current.Hero.LP <= 0 Then

                With Instances.Current
                    FightInfo = [String].Concat(.Hero.Name, "͐퓬s\ɊׂB")
                    With .Renderer
                        .RefreshScreen()
                        .DoAnimation(New String() {"HM 2", "WT 1000"})
                        .RenderGameOver()
                    End With
                    .ExitGame()
                End With

            ElseIf Enemy.LP <= 0 Then

                If GetType(Boss).IsInstanceOfType(Enemy) Then
                    Dim location As PointF = New PointF(0.0F, 120.0F)
                    For Each aMsg As String In CType(Enemy, Boss).GetFinishMsgs()
                        Instances.Current.Renderer.ShowMessage(aMsg, location, 34.0F, Color.Red, 2000)
                        location = New PointF(location.X + 78.0F, location.Y + 40.0F)
                    Next aMsg
                End If

                FightInfo = "ǦނɐI"
                With Instances.Current
                    .Renderer.RenderedObjectInFight = SceneRenderer.VisibilityInFight.NoEnemy
                    .Renderer.RefreshScreen()
                    .GameInfo.Money += Enemy.Money
                    .GameInfo.Win += 1
                    .Audio.Stop()
                    If .GameInfo.Win Mod 20 = 0 Then
                        With New LevelUp
                            .ShowDialog(Instances.Current.Form)
                            .Dispose()
                        End With
                    Else
                        .Wait(1000)
                    End If
                End With

                won = True
                Exit Do

            End If
        Loop

        Instances.Current.Renderer.RemoveRenderer(Me)

        smallFont.Dispose()
        mediumFont.Dispose()

    End Sub
End Class
