Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions

Public NotInheritable Class GameInfo

    Private _areaName As String = ""
    Private _location As Point = Point.Empty
    Private _breakthrough As Integer = 0
    Private _win As Integer = 0
    Private _fight As Integer = 0
    Private _money As Integer = 0
    Private _defeatedList As ArrayList = New ArrayList

    Public Property AreaName() As String
        Get
            Return _areaName
        End Get
        Set(ByVal Value As String)
            _areaName = Value
        End Set
    End Property

    Public Property Location() As Point
        Get
            Return _location
        End Get
        Set(ByVal Value As Point)
            _location = Value
        End Set
    End Property

    Public Property Breakthrough() As Integer
        Get
            Return _breakthrough
        End Get
        Set(ByVal Value As Integer)
            _breakthrough = Value
        End Set
    End Property

    Public Property Win() As Integer
        Get
            Return _win
        End Get
        Set(ByVal Value As Integer)
            _win = Value
        End Set
    End Property

    Public Property Fight() As Integer
        Get
            Return _fight
        End Get
        Set(ByVal Value As Integer)
            _fight = Value
        End Set
    End Property

    Public Property Money() As Integer
        Get
            Return _money
        End Get
        Set(ByVal Value As Integer)
            _money = Value
        End Set
    End Property

    Public Sub AddDefeated(ByVal bossName As String)
        _defeatedList.Add(bossName)
    End Sub

    Public Function IsDefeated(ByVal bossName As String) As Boolean
        Return _defeatedList.Contains(bossName)
    End Function

    Private Sub New()

    End Sub

    Public Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim newGameInfo As New GameInfo

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "(.+)=(.+)")
                        If .Success Then
                            Dim paramName As String = .Groups(1).Value.Trim()
                            Dim paramValue As String = .Groups(2).Value.Trim()
                            With newGameInfo
                                Select Case paramName.ToUpper()
                                    Case "AREANAME"
                                        ._areaName = paramValue
                                    Case "LOCATIONX"
                                        ._location = New Point([Integer].Parse(paramValue), ._location.Y)
                                    Case "LOCATIONY"
                                        ._location = New Point(._location.X, [Integer].Parse(paramValue))
                                    Case "BREAKTHROUGH"
                                        ._breakthrough = [Integer].Parse(paramValue)
                                    Case "WIN"
                                        ._win = [Integer].Parse(paramValue)
                                    Case "FIGHT"
                                        ._fight = [Integer].Parse(paramValue)
                                    Case "MONEY"
                                        ._money = [Integer].Parse(paramValue)
                                    Case "DEFEATEDBOSS"
                                        ._defeatedList.Add(paramValue)
                                End Select
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return newGameInfo

        Finally
            reader.Close()
        End Try

    End Function

    Public Sub SaveData(ByVal filePath As String)

        Dim theFile As New FileStream(filePath, FileMode.Create, FileAccess.Write, FileShare.None)
        Try
            SaveData(theFile)
        Finally
            theFile.Close()
        End Try

    End Sub

    Private Sub SaveData(ByVal theStream As Stream)

        Dim writer As New StreamWriter(theStream, GetINIStreamEncoding())
        Try
            writer.WriteLine([String].Format("AreaName={0}", _areaName))
            writer.WriteLine([String].Format("LocationX={0}", _location.X))
            writer.WriteLine([String].Format("LocationY={0}", _location.Y))
            writer.WriteLine([String].Format("Breakthrough={0}", _breakthrough))
            writer.WriteLine([String].Format("Win={0}", _win))
            writer.WriteLine([String].Format("Fight={0}", _fight))
            writer.WriteLine([String].Format("Money={0}", _money))
            For Each aBossName As String In _defeatedList
                writer.WriteLine([String].Format("DefeatedBoss={0}", aBossName))
            Next aBossName

        Finally
            writer.Close()
        End Try

    End Sub
End Class
