Imports System.Math
Imports System.IO
Imports System.Text.RegularExpressions

Public Class Hero
    Inherits Actor
    Implements IFighter

    Private _name As String = ""
    Private _physicalEnergy As Integer = 0
    Private _psychicEnergy As Integer = 0
    Private _strength As Integer = 0
    Private _quickness As Integer = 0
    Private _lifePoint As Integer = 0
    Private _techPoint As Integer = 0
    Private _weapon As Weapon = Nothing
    Private _guard As Guard = Nothing
    Private _tools As ItemList = New ItemList

    Private Shared _physicalEnergyBase As Integer = 0
    Public Shared ReadOnly Property PhysicalEnergyBase() As Integer
        Get
            Return _physicalEnergyBase
        End Get
    End Property
    Private Shared _psychicEnergyBase As Integer = 0
    Public Shared ReadOnly Property PsychicEnergyBase() As Integer
        Get
            Return _psychicEnergyBase
        End Get
    End Property
    Private Shared _strengthBase As Integer = 0
    Public Shared ReadOnly Property StrengthBase() As Integer
        Get
            Return _strengthBase
        End Get
    End Property
    Private Shared _quicknessBase As Integer = 0
    Public Shared ReadOnly Property QuicknessBase() As Integer
        Get
            Return _quicknessBase
        End Get
    End Property

    Public Property Name() As String
        Get
            Return _name
        End Get
        Set(ByVal Value As String)
            _name = Value
        End Set
    End Property

    Public Property PhysicalEnergy() As Integer
        Get
            Return _physicalEnergy
        End Get
        Set(ByVal Value As Integer)
            _physicalEnergy = Max(0, Value)
        End Set
    End Property

    Public ReadOnly Property MaxLP() As Integer Implements IFighter.MaxLP
        Get
            Return PhysicalEnergy * 10
        End Get
    End Property

    Public Property PsychicEnergy() As Integer
        Get
            Return _psychicEnergy
        End Get
        Set(ByVal Value As Integer)
            _psychicEnergy = Max(0, Value)
        End Set
    End Property

    Public ReadOnly Property MaxTP() As Integer Implements IFighter.MaxTP
        Get
            Return PsychicEnergy * 10
        End Get
    End Property

    Public Property Strength() As Integer
        Get
            Return _strength
        End Get
        Set(ByVal Value As Integer)
            _strength = Max(0, Value)
        End Set
    End Property

    Public Property Quickness() As Integer
        Get
            Return _quickness
        End Get
        Set(ByVal Value As Integer)
            _quickness = Max(0, Value)
        End Set
    End Property

    Public Property LP() As Integer Implements IFighter.LP
        Get
            Return _lifePoint
        End Get
        Set(ByVal Value As Integer)
            _lifePoint = Min(Max(0, Value), MaxLP)
        End Set
    End Property

    Public Property TP() As Integer Implements IFighter.TP
        Get
            Return _techPoint
        End Get
        Set(ByVal Value As Integer)
            _techPoint = Min(Max(0, Value), MaxTP)
        End Set
    End Property

    Public Property Weapon() As Weapon
        Get
            Return _weapon
        End Get
        Set(ByVal Value As Weapon)
            _weapon = Value
        End Set
    End Property

    Public Property Guard() As Guard
        Get
            Return _guard
        End Get
        Set(ByVal Value As Guard)
            _guard = Value
        End Set
    End Property

    Public ReadOnly Property OP() As Integer Implements IFighter.OP
        Get
            If Not Weapon Is Nothing Then
                Return Strength + Weapon.OP
            Else
                Return Strength
            End If
        End Get
    End Property

    Public ReadOnly Property DP() As Integer Implements IFighter.DP
        Get
            If Not Guard Is Nothing Then
                Return Guard.DP + Quickness \ 2
            Else
                Return Quickness \ 2
            End If
        End Get
    End Property

    Public ReadOnly Property MV() As Integer Implements IFighter.MV
        Get
            If Not Guard Is Nothing Then
                Return Max(1, Quickness - Guard.Weight)
            Else
                Return Quickness
            End If
        End Get
    End Property

    Public ReadOnly Property Tools() As ItemList
        Get
            Return _tools
        End Get
    End Property

    Public Overrides Function GetSkinBounds() As Rectangle

        Dim bounds As New Rectangle(0, 0, 32, 32)

        Select Case Direction
            Case Direction.Down
                bounds.Y = 256 + 32 * 0
            Case Direction.Up
                bounds.Y = 256 + 32 * 1
            Case Direction.Right
                bounds.Y = 256 + 32 * 2
            Case Direction.Left
                bounds.Y = 256 + 32 * 3
        End Select

        If Frames \ 2 = 0 Then
            bounds.X = 32 * 0
        Else
            bounds.X = 32 * 1
        End If

        Return bounds

    End Function

    Protected Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim newHero As New Hero("Hero")

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "(.+)=(.+)")
                        If .Success Then
                            Dim paramName As String = .Groups(1).Value.Trim()
                            Dim paramValue As String = .Groups(2).Value.Trim()
                            With newHero
                                Select Case paramName.ToUpper()
                                    Case "NAME"
                                        ._name = paramValue
                                    Case "PHYSICALENERGY"
                                        ._physicalEnergy = [Integer].Parse(paramValue)
                                        If _physicalEnergyBase = 0 Then
                                            _physicalEnergyBase = ._physicalEnergy
                                        End If
                                    Case "PSYCHICENERGY"
                                        ._psychicEnergy = [Integer].Parse(paramValue)
                                        If _psychicEnergyBase = 0 Then
                                            _psychicEnergyBase = ._psychicEnergy
                                        End If
                                    Case "STRENGTH"
                                        ._strength = [Integer].Parse(paramValue)
                                        If _strengthBase = 0 Then
                                            _strengthBase = ._strength
                                        End If
                                    Case "QUICKNESS"
                                        ._quickness = [Integer].Parse(paramValue)
                                        If _quicknessBase = 0 Then
                                            _quicknessBase = ._quickness
                                        End If
                                    Case "LP"
                                        ._lifePoint = [Integer].Parse(paramValue)
                                    Case "TP"
                                        ._techPoint = [Integer].Parse(paramValue)
                                    Case "WEAPON"
                                        ._weapon = CType(Instances.Current.GetItem(paramValue).Clone(), Weapon)
                                    Case "FORGED"
                                        ._weapon.Forged = [Integer].Parse(paramValue)
                                    Case "BROKEN"
                                        ._weapon.Broken = [Integer].Parse(paramValue)
                                    Case "GUARD"
                                        ._guard = CType(Instances.Current.GetItem(paramValue).Clone(), Guard)
                                    Case "TOOLS"
                                        ._tools.Add(Instances.Current.GetItem(paramValue))
                                End Select
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return newHero

        Finally
            reader.Close()
        End Try

    End Function

    Public Sub SaveData(ByVal filePath As String)

        Dim theFile As New FileStream(filePath, FileMode.Create, FileAccess.Write, FileShare.None)
        Try
            SaveData(theFile)
        Finally
            theFile.Close()
        End Try

    End Sub

    Private Sub SaveData(ByVal theStream As Stream)

        Dim writer As New StreamWriter(theStream, GetINIStreamEncoding())
        Try
            writer.WriteLine([String].Format("Name={0}", _name))
            writer.WriteLine([String].Format("PhysicalEnergy={0}", _physicalEnergy))
            writer.WriteLine([String].Format("PsychicEnergy={0}", _psychicEnergy))
            writer.WriteLine([String].Format("Strength={0}", _strength))
            writer.WriteLine([String].Format("Quickness={0}", _quickness))
            writer.WriteLine([String].Format("LP={0}", _lifePoint))
            writer.WriteLine([String].Format("TP={0}", _techPoint))
            writer.WriteLine([String].Format("Weapon={0}", _weapon.Tag))
            writer.WriteLine([String].Format("Forged={0}", _weapon.Forged))
            writer.WriteLine([String].Format("Broken={0}", _weapon.Broken))
            writer.WriteLine([String].Format("Guard={0}", _guard.Tag))
            For Each anItem As Item In _tools
                writer.WriteLine([String].Format("Tools={0}", anItem.Tag))
            Next anItem

        Finally
            writer.Close()
        End Try

    End Sub
End Class
