Imports GameLib

Public NotInheritable Class Instances
    Implements IDisposable

    Private Shared _current As Instances = Nothing

    Private _renderer As SceneRenderer = Nothing
    Private _audio As AudioPlayer = Nothing
    Private _fightManager As FightManager = Nothing
    Private _gameInfo As GameInfo = Nothing
    Private _hero As Hero = Nothing
    Private _currentArea As Area = Nothing
    Private _items As Hashtable = New Hashtable
    Private _bosses As Hashtable = New Hashtable

    Private _keyboard As KeyboardDevice = New KeyboardDevice

    Public ReadOnly Property Renderer() As SceneRenderer
        Get
            Return _renderer
        End Get
    End Property

    Public ReadOnly Property Audio() As AudioPlayer
        Get
            Return _audio
        End Get
    End Property

    Public ReadOnly Property Fight() As FightManager
        Get
            Return _fightManager
        End Get
    End Property

    Public ReadOnly Property GameInfo() As GameInfo
        Get
            Return _gameInfo
        End Get
    End Property

    Public ReadOnly Property Hero() As Hero
        Get
            Return _hero
        End Get
    End Property

    Public ReadOnly Property CurrentArea() As Area
        Get
            Return _currentArea
        End Get
    End Property

    Public Function GetItem(ByVal tagName As String) As Item

        Return CType(_items(tagName), Item)

    End Function

    Public Shared ReadOnly Property Current() As Instances
        Get
            Return _current
        End Get
    End Property

    Private Sub Hero_Moved(ByVal sender As Object, ByVal e As ActorMovedEventArgs)

        Dim thereIsActor As Boolean = False

        GameInfo.Location = Area.WorldCoordToAreaCoord(Hero.Bounds)

        For Each anActor As Actor In CurrentArea.GetActors()
            With anActor
                If Not .Bounds.Contains(e.PreviousBounds) AndAlso .Bounds.IntersectsWith(e.PreviousBounds) Then
                    If .Bounds.Contains(e.CurrentBounds) Then
                        If .Interrupt(Action.Interrupting.Enter, Nothing) = True Then
                            thereIsActor = True
                        End If
                    ElseIf Not .Bounds.IntersectsWith(e.CurrentBounds) Then
                        .Interrupt(Action.Interrupting.Leave, Nothing)
                    End If
                End If
            End With
        Next anActor

        If e.CurrentBounds.X Mod Area.ChipWidth = 0 AndAlso e.CurrentBounds.Y Mod Area.ChipHeight = 0 Then
            If Not Form Is Nothing Then
                'Form.Text = [String].Format("RPG - {0}", GameInfo.Location)
            End If
            If Not thereIsActor Then
                If CurrentArea.IsMaze Then
                    Static randomizer As Random = New Random
                    Static nextEncounter As Integer = 20
                    Static currStep As Integer = 0
                    currStep += 1
                    If currStep = nextEncounter Then
                        DoFight(New Enemy(randomizer.Next(0, 8)))
                        nextEncounter = randomizer.Next(15, 25)
                        currStep = 0
                    End If
                End If
            End If
        End If

    End Sub

    Private Function LoadData() As Boolean

        Try
            _gameInfo = CType(RPG.GameInfo.LoadData("SavedGameInfo.ini"), GameInfo)
            If Not _hero Is Nothing Then
                RemoveHandler _hero.Moved, AddressOf Hero_Moved
            End If
            _hero = CType(RPG.Hero.LoadData("SavedHero.ini"), Hero)
            AddHandler _hero.Moved, AddressOf Hero_Moved

            _hero.Bounds = Area.AreaCoordToWorldCoord(_gameInfo.Location)
            _currentArea = CType(Area.LoadData([String].Concat(_gameInfo.AreaName, ".ini")), Area)
            _currentArea.AddActor(_hero)

            Return True
        Catch
            Return False
        End Try

    End Function

    Public Function SaveData() As Boolean

        Try
            GameInfo.SaveData("SavedGameInfo.ini")
            Hero.SaveData("SavedHero.ini")
            Return True
        Catch
            Return False
        End Try

    End Function

    Private Sub New()

        Try
            _renderer = New SceneRenderer
            _audio = New AudioPlayer
            _audio.LoadBGMConfig("BGM.ini")

            _gameInfo = CType(RPG.GameInfo.LoadData("GameInfo.ini"), GameInfo)

            Dim healers() As LPHealer = CType(LPHealer.LoadData("LPHealer.ini"), LPHealer())
            Dim bombs() As Bomb = CType(Bomb.LoadData("Bomb.ini"), Bomb())
            Dim mappers() As Mapper = CType(Mapper.LoadData("Mapper.ini"), Mapper())
            Dim weapons() As Weapon = CType(Weapon.LoadData("Weapon.ini"), Weapon())
            For Each aWeapon As Weapon In weapons
                aWeapon.LoadTechMotions()
            Next aWeapon
            Dim guards() As Guard = CType(Guard.LoadData("Guard.ini"), Guard())

            With New ArrayList
                .AddRange(healers)
                .AddRange(bombs)
                .AddRange(mappers)
                .AddRange(weapons)
                .AddRange(guards)
                For Each anItem As Item In .ToArray(GetType(Item))
                    _items.Add(anItem.Tag, anItem)
                Next anItem
            End With

            For Each aBoss As Boss In CType(Boss.LoadData("Boss.ini"), Boss())
                _bosses.Add(aBoss.Tag, aBoss)
            Next aBoss

            _hero = CType(RPG.Hero.LoadData("Hero.ini"), Hero)
            _hero.Bounds = Area.AreaCoordToWorldCoord(_gameInfo.Location)
            _currentArea = CType(Area.LoadData([String].Concat(_gameInfo.AreaName, ".ini")), Area)
            _currentArea.AddActor(_hero)

            AddHandler _hero.Moved, AddressOf Hero_Moved

            mainScreen = New RPGMainForm
            With mainScreen
                .Text = Application.ProductName
                .MaximizeBox = False
                .FormBorderStyle = FormBorderStyle.FixedSingle
                .ClientSize = New Size(640, 480)
                .StartPosition = FormStartPosition.CenterScreen
                .Show()
                .Activate()
            End With

            _audio.Play(AudioPlayer.BGMNameInTown)

            _renderer.SetTargetWindow(mainScreen)
            _keyboard.SetInputTarget(mainScreen)

        Catch

            Dispose(True)
            Throw

        End Try

    End Sub

    Private disposed As Boolean = False

    Public ReadOnly Property IsDisposed() As Boolean
        Get
            Return disposed
        End Get
    End Property

    Private Sub Dispose(ByVal disposing As Boolean)

        If Not IsDisposed Then

            If disposing Then

                If Not _hero Is Nothing Then
                    RemoveHandler _hero.Moved, AddressOf Hero_Moved
                End If

                If Not _audio Is Nothing Then
                    _audio.Dispose()
                    _audio = Nothing
                End If

                If Not _renderer Is Nothing Then
                    _renderer.Dispose()
                    _renderer = Nothing
                End If

                If Not mainScreen Is Nothing Then
                    mainScreen.Dispose()
                    mainScreen = Nothing
                End If

            End If

        End If

        disposed = True

    End Sub

    Private Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
        Dispose(False)
    End Sub

    Private Class RPGMainForm
        Inherits Form

        Protected Overrides Sub OnClosing(ByVal e As System.ComponentModel.CancelEventArgs)
            MyBase.OnClosing(e)

            Instances.Current.breaking = True
            e.Cancel = True

        End Sub

        Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)

            Dim renderer As SceneRenderer = Instances.Current.Renderer
            If Not renderer Is Nothing AndAlso Not renderer.IsDisposed Then
                renderer.RefreshScreen()
            End If

        End Sub
    End Class

    Private mainScreen As Form = Nothing
    Private breaking As Boolean = False

    Public ReadOnly Property Form() As Form
        Get
            Return mainScreen
        End Get
    End Property

    Public Sub ChangeArea(ByVal areaName As String, ByVal location As Point)

        Renderer.FadeOut()
        _currentArea = CType(Area.LoadData([String].Concat(areaName, ".ini")), Area)
        _currentArea.AddActor(_hero)
        _gameInfo.AreaName = areaName
        _hero.Bounds = Area.AreaCoordToWorldCoord(location)
        _hero.ClearInstantActions()
        _hero.Frames = 0
        Renderer.FadeIn()

        Audio.Play(AudioPlayer.BGMNameInTown)

    End Sub

    Public Sub ChangeArea(ByVal basement As Integer, ByVal down As Boolean)

        Renderer.FadeOut()
        Dim startLocation As Point = Point.Empty
        _currentArea = Area.CreateMaze(basement, startLocation, down)
        _currentArea.AddActor(_hero)
        _gameInfo.AreaName = Nothing
        _hero.Bounds = Area.AreaCoordToWorldCoord(startLocation)
        _hero.ClearInstantActions()
        _hero.Frames = 0
        Renderer.FadeIn()

        Audio.Play(AudioPlayer.BGMNameInDungeon)

    End Sub

    Public Function DoFight(ByVal bossName As String) As Boolean

        Return DoFight(CType(_bosses(bossName), Boss))

    End Function

    Public Function DoFight(ByVal theEnemy As Enemy) As Boolean

        Dim prevMusicName As String = Audio.CurrentMusicName

        If GetType(Boss).IsInstanceOfType(theEnemy) Then
            Dim location As PointF = New PointF(0.0F, 280.0F)
            For Each aMsg As String In CType(theEnemy, Boss).GetBeginMsgs()
                Renderer.ShowMessage(aMsg, location, 34.0F, Color.LightCoral, 2000)
                location = New PointF(location.X + 78.0F, location.Y + 40.0F)
            Next aMsg
        Else
            Audio.Stop()
        End If

        Renderer.FlashScreen(1000)

        _fightManager = New FightManager(theEnemy)

        If Not GetType(Boss).IsInstanceOfType(theEnemy) Then
            Audio.Play(AudioPlayer.BGMNameInNormalFight)
        Else
            Audio.Play(CType(theEnemy, Boss).BGMName)
        End If

        _fightManager.Fire()
        Dim won As Boolean = _fightManager.HeroWon

        Audio.Play(prevMusicName)

        _fightManager = Nothing

        Return won

    End Function

    Public ReadOnly Property IsFighting() As Boolean
        Get
            If Fight Is Nothing Then
                Return False
            Else
                Return True
            End If
        End Get
    End Property

    Public Sub [ExitGame]()

        Throw New ApplicationException("Current game has been closed")

    End Sub

    Public Sub Wait(ByVal timeout As Integer)
        Wait(TimeSpan.FromMilliseconds(timeout))
    End Sub

    Public Sub Wait(ByVal timeout As TimeSpan)

        Dim t0 As Date = [Date].Now
        Do
            Application.DoEvents()

            Dim dt As TimeSpan = [Date].Now.Subtract(t0)
            If dt.CompareTo(timeout) >= 0 Then
                Exit Do
            Else
                Threading.Thread.Sleep(10)
            End If

            If breaking Then
                ExitGame()
            End If
        Loop

    End Sub

    Private Sub DoMainLoop()

        Application.DoEvents()

        Dim cycleTime As TimeSpan = TimeSpan.FromMilliseconds(50)

        Do Until breaking

            Dim t0 As Date = [Date].Now

            If Not Hero.HasInstantAction() AndAlso Not CurrentArea.IsAnyoneBusy() Then
                If _keyboard.IsDown(Keys.Right) Then
                    Hero.AddInstantAction("10 STEPRIGHT")
                ElseIf _keyboard.IsDown(Keys.Left) Then
                    Hero.AddInstantAction("10 STEPLEFT")
                ElseIf _keyboard.IsDown(Keys.Down) Then
                    Hero.AddInstantAction("10 STEPDOWN")
                ElseIf _keyboard.IsDown(Keys.Up) Then
                    Hero.AddInstantAction("10 STEPUP")
                ElseIf _keyboard.IsDown(Keys.Enter, True) Then
                    Dim bounds As Rectangle = Hero.Bounds
                    Dim theActor As Actor = CurrentArea.GetActor(Hero, bounds)
                    If theActor Is Nothing OrElse theActor.Interrupt(Action.Interrupting.Contact, Nothing) = False Then
                        Select Case Hero.Direction
                            Case Direction.Right
                                bounds.Offset(32, 0)
                            Case Direction.Left
                                bounds.Offset(-32, 0)
                            Case Direction.Down
                                bounds.Offset(0, 32)
                            Case Direction.Up
                                bounds.Offset(0, -32)
                        End Select
                        theActor = CurrentArea.GetActor(Hero, bounds)
                        If Not theActor Is Nothing Then
                            theActor.Interrupt(Action.Interrupting.Contact, Nothing)
                        End If
                    End If
                ElseIf _keyboard.IsDown(Keys.Space, True) Then
                    With New Status
                        .ShowDialog(Form)
                        .Dispose()
                    End With
                End If
            End If

            For Each anActor As Actor In CurrentArea.GetActors()
                anActor.Proceed()
            Next anActor

            Renderer.RefreshScreen()

            Dim dt As TimeSpan = [Date].Now.Subtract(t0)
            If cycleTime.CompareTo(dt) > 0 Then
                Wait(cycleTime.Subtract(dt))
            Else
                Wait(0)
            End If
        Loop

    End Sub

    Public Shared Sub Fire(ByVal continue As Boolean)

        Try
            If continue Then
                _current = New Instances
                If _current.LoadData() = False Then
                    Trace.Fail("f[^̓ǂݍ݂Ɏs܂B")
                    Exit Sub
                End If
            Else
                Dim heroName As String = InputBox("v[[̖O͂ĂB", "New Game", "Player")
                If heroName <> "" Then
                    _current = New Instances
                    _current._hero.Name = heroName
                Else
                    Exit Sub
                End If
            End If
            _current.DoMainLoop()
        Catch ex As ApplicationException
            'Ignore
        Catch ex As Exception
            Trace.Fail(ex.ToString())
        Finally
            If Not _current Is Nothing Then
                _current.Dispose()
                _current = Nothing
            End If
        End Try

    End Sub
End Class
