Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions

Public MustInherit Class Item

    Private _tag As String = ""
    Private _name As String = ""
    Private _comment As String = ""
    Private _price As Integer = 0

    Protected Overridable Sub CopyTo(ByVal dest As Item)

        With dest
            ._tag = _tag
            ._name = _name
            ._comment = Comment
            ._price = _price
        End With

    End Sub

    Public MustOverride Function Clone() As Item

    Public ReadOnly Property Tag() As String
        Get
            Return _tag
        End Get
    End Property

    Public Property Name() As String
        Get
            Return _name
        End Get
        Set(ByVal Value As String)
            _name = Value
        End Set
    End Property

    Public Property Comment() As String
        Get
            Return _comment
        End Get
        Set(ByVal Value As String)
            _comment = Value
        End Set
    End Property

    Public Property Price() As Integer
        Get
            Return _price
        End Get
        Set(ByVal Value As Integer)
            _price = Value
        End Set
    End Property

    Private Sub New()

    End Sub

    Protected Sub New(ByVal tagName As String)

        _tag = tagName

    End Sub

    Protected Function LoadData(ByVal paramName As String, ByVal paramValue As String) As Boolean

        Select Case paramName.ToUpper()
            Case "NAME"
                _name = paramValue
            Case "COMMENT"
                _comment = paramValue
            Case "PRICE"
                _price = [Integer].Parse(paramValue)
            Case Else
                Return False
        End Select

        Return True

    End Function
End Class

Public Class ItemList
    Inherits CollectionBase

    Default Public Property Item(ByVal index As Integer) As Item
        Get
            Return CType(Me.InnerList.Item(index), Item)
        End Get
        Set(ByVal Value As Item)
            Me.InnerList.Item(index) = Value
        End Set
    End Property

    Public Sub Add(ByVal value As Item)

        Me.InnerList.Add(value)

    End Sub

    Public Sub Add(ByVal value As Array)

        For Each anItem As Item In value
            Me.InnerList.Add(anItem)
        Next anItem

    End Sub
End Class

Public Class LPHealer
    Inherits Item

    Private _recovery As Integer = 0

    Protected Overrides Sub CopyTo(ByVal dest As Item)
        MyBase.CopyTo(dest)

        With CType(dest, LPHealer)
            ._recovery = _recovery
        End With

    End Sub

    Public Overrides Function Clone() As Item

        Dim dest As New LPHealer(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Property Recovery() As Integer
        Get
            Return _recovery
        End Get
        Set(ByVal Value As Integer)
            _recovery = Value
        End Set
    End Property

    Public Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim healerList As New ArrayList
        Dim newHealer As LPHealer = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newHealer = New LPHealer(.Groups(1).Value.Trim())
                            healerList.Add(newHealer)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newHealer
                                        If Not .LoadData(paramName, paramValue) Then
                                            Select Case paramName.ToUpper()
                                                Case "RECOVERY"
                                                    ._recovery = [Integer].Parse(paramValue)
                                            End Select
                                        End If
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(healerList.ToArray(GetType(LPHealer)), LPHealer())

        Finally
            reader.Close()
        End Try

    End Function
End Class

Public Class Bomb
    Inherits Item

    Private _damage As Integer = 0
    Private _accidentality As Integer = 0

    Protected Overrides Sub CopyTo(ByVal dest As Item)
        MyBase.CopyTo(dest)

        With CType(dest, Bomb)
            ._damage = _damage
            ._accidentality = _accidentality
        End With

    End Sub

    Public Overrides Function Clone() As Item

        Dim dest As New Bomb(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Property Damage() As Integer
        Get
            Return _damage
        End Get
        Set(ByVal Value As Integer)
            _damage = Value
        End Set
    End Property

    Public Property Accidentality() As Integer
        Get
            Return _accidentality
        End Get
        Set(ByVal Value As Integer)
            _accidentality = Value
        End Set
    End Property

    Public Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim bombList As New ArrayList
        Dim newBomb As Bomb = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newBomb = New Bomb(.Groups(1).Value.Trim())
                            bombList.Add(newBomb)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newBomb
                                        If Not .LoadData(paramName, paramValue) Then
                                            Select Case paramName.ToUpper()
                                                Case "DAMAGE"
                                                    ._damage = [Integer].Parse(paramValue)
                                                Case "ACCIDENTALITY"
                                                    ._accidentality = [Integer].Parse(paramValue)
                                            End Select
                                        End If
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(bombList.ToArray(GetType(Bomb)), Bomb())

        Finally
            reader.Close()
        End Try

    End Function
End Class

Public Class Mapper
    Inherits Item

    Private _isPositioningHero As Boolean

    Protected Overrides Sub CopyTo(ByVal dest As Item)
        MyBase.CopyTo(dest)

        With CType(dest, Mapper)
            ._isPositioningHero = _isPositioningHero
        End With

    End Sub

    Public Overrides Function Clone() As Item

        Dim dest As New Mapper(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Property IsPositioningHero() As Boolean
        Get
            Return _isPositioningHero
        End Get
        Set(ByVal Value As Boolean)
            _isPositioningHero = Value
        End Set
    End Property

    Public Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim mapperList As New ArrayList
        Dim newMapper As Mapper = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newMapper = New Mapper(.Groups(1).Value.Trim())
                            mapperList.Add(newMapper)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newMapper
                                        If Not .LoadData(paramName, paramValue) Then
                                            Select Case paramName.ToUpper()
                                                Case "POSITIONING"
                                                    .IsPositioningHero = [Boolean].Parse(paramValue)
                                            End Select
                                        End If
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(mapperList.ToArray(GetType(Mapper)), Mapper())

        Finally
            reader.Close()
        End Try

    End Function
End Class

Public Structure WeaponTech

    Private _name As String
    Private _mulOP As Integer
    Private _mulDP As Integer
    Private motions() As String

    Public Function Clone() As WeaponTech

        Dim dest As New WeaponTech

        With dest
            ._name = _name
            ._mulOP = _mulOP
            ._mulDP = _mulDP
            .motions = CType(motions.Clone(), String())
        End With

        Return dest

    End Function

    Public Sub New(ByVal name As String, ByVal mulOP As Integer, ByVal mulDP As Integer)

        With Me
            ._name = name
            ._mulOP = mulOP
            ._mulDP = mulDP
        End With

    End Sub

    Public Property Name() As String
        Get
            Return _name
        End Get
        Set(ByVal Value As String)
            _name = Value
        End Set
    End Property

    Public Property MulOP() As Integer
        Get
            Return _mulOP
        End Get
        Set(ByVal Value As Integer)
            _mulOP = Value
        End Set
    End Property

    Public Property MulDP() As Integer
        Get
            Return _mulDP
        End Get
        Set(ByVal Value As Integer)
            _mulDP = Value
        End Set
    End Property

    Public Function GetMotions() As String()
        Return CType(motions.Clone(), String())
    End Function

    Public Sub LoadMotions(ByVal filePath As String)

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            LoadMotions(theFile)
        Finally
            theFile.Close()
        End Try

    End Sub

    Public Sub LoadMotions(ByVal theStream As Stream)

        Dim lines As New ArrayList

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                lines.Add(line)
                line = reader.ReadLine()
            Loop

            motions = CType(lines.ToArray(GetType(String)), String())

        Finally
            reader.Close()
        End Try

    End Sub
End Structure

Public Class Weapon
    Inherits Item

    Private _baseOP As Integer = 0
    Private _maxOP As Integer = 0
    Private _forged As Integer = 0
    Private _broken As Integer = 0

    Private _tech1 As WeaponTech = New WeaponTech("", 100, 100)
    Private _tech2 As WeaponTech = New WeaponTech("", 100, 100)

    Protected Overrides Sub CopyTo(ByVal dest As Item)
        MyBase.CopyTo(dest)

        With CType(dest, Weapon)
            ._baseOP = _baseOP
            ._maxOP = _maxOP
            ._forged = _forged
            ._broken = _broken
            ._tech1 = _tech1.Clone()
            ._tech2 = _tech2.Clone()
        End With

    End Sub

    Public Overrides Function Clone() As Item

        Dim dest As New Weapon(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Property BaseOP() As Integer
        Get
            Return _baseOP
        End Get
        Set(ByVal Value As Integer)
            _baseOP = Value
        End Set
    End Property

    Public Property MaxOP() As Integer
        Get
            Return _maxOP
        End Get
        Set(ByVal Value As Integer)
            _maxOP = Value
        End Set
    End Property

    Public Property Forged() As Integer
        Get
            Return _forged
        End Get
        Set(ByVal Value As Integer)
            _forged = Min(Value, MaxOP - BaseOP)
        End Set
    End Property

    Public Property Broken() As Integer
        Get
            Return _broken
        End Get
        Set(ByVal Value As Integer)
            _broken = Min(Value, BaseOP + Forged)
        End Set
    End Property

    Public ReadOnly Property CanUseTech2() As Boolean
        Get
            If BaseOP + Forged = MaxOP Then
                Return True
            Else
                Return False
            End If
        End Get
    End Property

    Public ReadOnly Property OP() As Integer
        Get
            Return Max(0, BaseOP + Forged - Broken)
        End Get
    End Property

    Public Property Tech1() As WeaponTech
        Get
            Return _tech1
        End Get
        Set(ByVal Value As WeaponTech)
            _tech1 = Value
        End Set
    End Property

    Public Property Tech2() As WeaponTech
        Get
            Return _tech2
        End Get
        Set(ByVal Value As WeaponTech)
            _tech2 = Value
        End Set
    End Property

    Protected Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Sub LoadTechMotions()

        _tech1.LoadMotions(Path.Combine("TechMotions", [String].Concat(_tech1.Name, ".ini")))
        _tech2.LoadMotions(Path.Combine("TechMotions", [String].Concat(_tech2.Name, ".ini")))

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim weaponList As New ArrayList
        Dim newWeapon As Weapon = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newWeapon = New Weapon(.Groups(1).Value.Trim())
                            weaponList.Add(newWeapon)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newWeapon
                                        If Not .LoadData(paramName, paramValue) Then
                                            Select Case paramName.ToUpper()
                                                Case "BASEOP"
                                                    ._baseOP = [Integer].Parse(paramValue)
                                                Case "MAXOP"
                                                    ._maxOP = [Integer].Parse(paramValue)
                                                Case "FORGED"
                                                    ._forged = [Integer].Parse(paramValue)
                                                Case "BROKEN"
                                                    ._broken = [Integer].Parse(paramValue)
                                                Case "TECH1_NAME"
                                                    ._tech1.Name = paramValue
                                                Case "TECH1_MULOP"
                                                    ._tech1.MulOP = [Integer].Parse(paramValue)
                                                Case "TECH1_MULDP"
                                                    ._tech1.MulDP = [Integer].Parse(paramValue)
                                                Case "TECH2_NAME"
                                                    ._tech2.Name = paramValue
                                                Case "TECH2_MULOP"
                                                    ._tech2.MulOP = [Integer].Parse(paramValue)
                                                Case "TECH2_MULDP"
                                                    ._tech2.MulDP = [Integer].Parse(paramValue)
                                            End Select
                                        End If
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(weaponList.ToArray(GetType(Weapon)), Weapon())

        Finally
            reader.Close()
        End Try

    End Function
End Class

Public Class Guard
    Inherits Item

    Private _deffensePoint As Integer = 0
    Private _weight As Integer = 0

    Protected Overrides Sub CopyTo(ByVal dest As Item)
        MyBase.CopyTo(dest)

        With CType(dest, Guard)
            ._deffensePoint = _deffensePoint
            ._weight = _weight
        End With

    End Sub

    Public Overrides Function Clone() As Item

        Dim dest As New Guard(Me.Tag)
        Me.CopyTo(dest)

        Return dest

    End Function

    Public Property DP() As Integer
        Get
            Return _deffensePoint
        End Get
        Set(ByVal Value As Integer)
            _deffensePoint = Value
        End Set
    End Property

    Public Property Weight() As Integer
        Get
            Return _weight
        End Get
        Set(ByVal Value As Integer)
            _weight = Value
        End Set
    End Property

    Protected Sub New(ByVal tagName As String)
        MyBase.New(tagName)

    End Sub

    Public Overloads Shared Function LoadData(ByVal filePath As String) As Object

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            Return LoadData(theFile)
        Finally
            theFile.Close()
        End Try

    End Function

    Private Overloads Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim guardList As New ArrayList
        Dim newGuard As Guard = Nothing

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = line.Trim()
                If Not line.StartsWith("#") Then
                    With Regex.Match(line, "\[(\w+)\]")
                        If .Success Then
                            newGuard = New Guard(.Groups(1).Value.Trim())
                            guardList.Add(newGuard)
                        Else
                            With Regex.Match(line, "(.+)=(.+)")
                                If .Success Then
                                    Dim paramName As String = .Groups(1).Value.Trim()
                                    Dim paramValue As String = .Groups(2).Value.Trim()
                                    With newGuard
                                        If Not .LoadData(paramName, paramValue) Then
                                            Select Case paramName.ToUpper()
                                                Case "DP"
                                                    ._deffensePoint = [Integer].Parse(paramValue)
                                                Case "WEIGHT"
                                                    ._weight = [Integer].Parse(paramValue)
                                            End Select
                                        End If
                                    End With
                                End If
                            End With
                        End If
                    End With
                End If
                line = reader.ReadLine()
            Loop

            Return CType(guardList.ToArray(GetType(Guard)), Guard())

        Finally
            reader.Close()
        End Try

    End Function
End Class
