Imports GameLib
Imports System.Text.RegularExpressions

Public Interface IAdditionalRenderer
    Sub Render(ByVal g As Graphics)
End Interface

Public NotInheritable Class SceneRenderer
    Implements IDisposable

    Private disposed As Boolean = False

    Private graphicsDev As GraphicDevice = Nothing
    Private backBuffer As RenderTarget = Nothing
    Private workBuffer As RenderTarget = Nothing
    Private texture As texture = Nothing
    Private additionalRenderer As ArrayList = New ArrayList

    Public Sub AddRenderer(ByVal renderer As IAdditionalRenderer)
        additionalRenderer.Add(renderer)
    End Sub

    Public Sub RemoveRenderer(ByVal renderer As IAdditionalRenderer)
        additionalRenderer.Remove(renderer)
    End Sub

    Public ReadOnly Property IsDisposed() As Boolean
        Get
            Return disposed
        End Get
    End Property

    Public Sub SetTargetWindow(ByVal Value As Control)
        graphicsDev.SetPresentTarget(Value)
    End Sub

    Public Sub New()

        graphicsDev = New GraphicDevice

        backBuffer = graphicsDev.CreateRenderTarget(640, 480)
        workBuffer = graphicsDev.CreateRenderTarget(640, 480)
        texture = graphicsDev.CreateTexture("MapChip.bmp", Color.White)

        graphicsDev.SetDestination(workBuffer)
        graphicsDev.PrepareSource(texture)

    End Sub

    Private Sub Dispose(ByVal disposing As Boolean)

        If Not IsDisposed Then

            If disposing Then

                If Not texture Is Nothing Then
                    graphicsDev.UnprepareSource()

                    texture.Dispose()
                    texture = Nothing
                End If

                If Not workBuffer Is Nothing Then
                    workBuffer.Dispose()
                    workBuffer = Nothing
                End If

                If Not backBuffer Is Nothing Then
                    backBuffer.Dispose()
                    backBuffer = Nothing
                End If

                If Not graphicsDev Is Nothing Then
                    graphicsDev.Dispose()
                    graphicsDev = Nothing
                End If

            End If

        End If

        disposed = True

    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overrides Sub Finalize()
        Dispose(False)
    End Sub

    'src Ŏw肳ꂽ̈̉摜Abounds Ŏw肳ꂽ̈Ƀ^Cɕ~l߂܂B
    '傫͕ύX܂B傤ǎ܂Ȃꍇ́A]؂̂Ă܂B
    ']͏ texture LoXŁA] target Ŏw肳ꂽLoXłB
    'transparent  True ̏ꍇAtexture  Color.White ɓ𓧉߂܂B
    '
    Private Sub TileBlt(ByVal src As Rectangle, ByVal bounds As Rectangle)

        Dim screenBounds As Rectangle = New Rectangle(Point.Empty, workBuffer.GetSize())

        Dim destBounds As New Rectangle(bounds.X, bounds.Y, 32, 32)

        Dim modX As Integer = bounds.Width Mod 32
        Dim modY As Integer = bounds.Height Mod 32

        graphicsDev.BeginScene()

        For y As Integer = bounds.Top To bounds.Bottom - 1 Step 32
            For x As Integer = bounds.Left To bounds.Right - 1 Step 32
                If screenBounds.IntersectsWith(destBounds) Then
                    graphicsDev.Render(destBounds, src)
                End If
                destBounds.Offset(32, 0)
            Next x
            If modX <> 0 Then
                src.Width = modX
                destBounds.Width = modX
                If screenBounds.IntersectsWith(destBounds) Then
                    graphicsDev.Render(destBounds, src)
                End If
                destBounds.Width = 32
                src.Width = 32
            End If
            destBounds.X = bounds.X
            destBounds.Offset(0, 32)
        Next y
        If modY <> 0 Then
            src.Height = modY
            destBounds.Height = modY
            For x As Integer = bounds.Left To bounds.Right - 1 Step 32
                If screenBounds.IntersectsWith(destBounds) Then
                    graphicsDev.Render(destBounds, src)
                End If
                destBounds.Offset(32, 0)
            Next x
            If modX <> 0 Then
                src.Width = modX
                destBounds.Width = modX
                If screenBounds.IntersectsWith(destBounds) Then
                    graphicsDev.Render(destBounds, src)
                End If
            End If
        End If

        graphicsDev.EndScene()

    End Sub

    'ʂobN obt@ɕ`悵܂B
    Private Sub Render()

        graphicsDev.Fill(workBuffer, Color.Black)

        If Not Instances.Current.IsFighting Then

            Dim screenBounds As Rectangle = New Rectangle(Point.Empty, workBuffer.GetSize())

            Dim data() As Area.AreaData = Instances.Current.CurrentArea.GetData()
            For Each aData As Area.AreaData In data
                Dim bounds As Rectangle = aData.Bounds
                bounds.Offset(-(Instances.Current.Hero.Bounds.X) + screenBounds.Width \ 2, -(Instances.Current.Hero.Bounds.Y + 16) + screenBounds.Height \ 2)
                TileBlt(New Rectangle((aData.ChipNumber Mod 16) * 32, (aData.ChipNumber \ 16) * 32, 32, 32), bounds)
            Next aData

            For Each anActor As Actor In Instances.Current.CurrentArea.GetActors()
                Dim bounds As Rectangle = anActor.Bounds
                bounds.Offset(-(Instances.Current.Hero.Bounds.X) + screenBounds.Width \ 2, -(Instances.Current.Hero.Bounds.Y + 16) + screenBounds.Height \ 2)
                TileBlt(anActor.GetSkinBounds(), bounds)
            Next anActor

        Else

            If (RenderedObjectInFight And VisibilityInFight.Hero) <> 0 Then
                RenderHero(HeroMotion.Normal)
            End If
            If (RenderedObjectInFight And VisibilityInFight.Enemy) <> 0 Then
                RenderEnemy()
            End If
            If (RenderedObjectInFight And VisibilityInFight.Others) <> 0 Then
                Dim g As Graphics = workBuffer.GetGraphics()
                For Each renderer As IAdditionalRenderer In additionalRenderer
                    renderer.Render(g)
                Next renderer
                workBuffer.ReleaseGraphics(g)
            End If

        End If

    End Sub

    Private ReadOnly heroBounds As Rectangle = New Rectangle(80, 256, 64, 64)
    Private ReadOnly enemyBounds As Rectangle = New Rectangle(400, 192, 128, 128)

    Public Enum HeroMotion
        Normal = 0
        EndSwing = 1
        Down = 2
        Guard = 3
        Blowing = 4
        BeginSwing = 5
    End Enum

    Public Enum VisibilityInFight
        [Nothing] = 0
        Hero = 1
        Enemy = 2
        Others = 4
        NoEnemy = Hero Or Others
        Everything = Hero Or Enemy Or Others
    End Enum

    Private _visibilityInFight As VisibilityInFight = VisibilityInFight.Everything
    Public Property RenderedObjectInFight() As VisibilityInFight
        Get
            Return _visibilityInFight
        End Get
        Set(ByVal Value As VisibilityInFight)
            _visibilityInFight = Value
        End Set
    End Property

    'l̃OtBbNX`悵܂B
    Private Sub RenderHero(ByVal motion As HeroMotion)

        graphicsDev.BeginScene()
        graphicsDev.Render(heroBounds, New Rectangle(64 * motion, 160, 64, 64))
        graphicsDev.EndScene()

    End Sub

    Private _enemyVisible As Boolean = False
    Public Property EnemyVisible() As Boolean
        Get
            Return _enemyVisible
        End Get
        Set(ByVal Value As Boolean)

        End Set
    End Property

    'G̃OtBbNX`悵܂B
    Private Sub RenderEnemy()

        graphicsDev.BeginScene()
        graphicsDev.Render(enemyBounds, Instances.Current.Fight.Enemy.GetSkinBounds())
        graphicsDev.EndScene()

    End Sub

    'motions Ŏw肳ꂽAj[V`悵܂B
    Public Sub DoAnimation(ByVal motions() As String)

        allowRefreshing = False

        For Each aMotion As String In motions

            Dim command As String = ""
            Dim params() As String = New String() {}
            With Regex.Match(aMotion.Trim(), "(\w\w)\s(.*)")
                If .Success Then
                    command = .Groups(1).Value
                    params = Regex.Replace(.Groups(2).Value, "\s", "").Split(","c)
                End If
            End With

            Select Case command.ToUpper()
                Case "FL"
                    graphicsDev.Fill(backBuffer, Color.White)
                    FlipToScreen()
                    Instances.Current.Wait([Integer].Parse(params(0)))
                Case "FC"
                    Dim r As Integer = [Integer].Parse(params(1))
                    Dim g As Integer = [Integer].Parse(params(2))
                    Dim b As Integer = [Integer].Parse(params(3))
                    graphicsDev.Fill(backBuffer, Color.FromArgb(r, g, b))
                    FlipToScreen()
                    Instances.Current.Wait([Integer].Parse(params(0)))
                Case "WT"
                    Instances.Current.Wait([Integer].Parse(params(0)))
                Case "DE"
                    Dim texnum As Integer = [Integer].Parse(params(0))
                    Dim x As Integer = [Integer].Parse(params(1))
                    Dim y As Integer = [Integer].Parse(params(2))
                    Dim destBounds As Rectangle = enemyBounds
                    destBounds.Offset(x, y)
                    Dim srcBounds As New Rectangle(texnum * 64, 32, 64, 64)
                    graphicsDev.BeginScene()
                    graphicsDev.Render(destbounds, srcbounds)
                    graphicsDev.EndScene()
                    FlipToBackBuffer()
                Case "DH"
                    Dim texnum As Integer = [Integer].Parse(params(0))
                    Dim x As Integer = [Integer].Parse(params(1))
                    Dim y As Integer = [Integer].Parse(params(2))
                    Dim destBounds As Rectangle = heroBounds
                    destBounds.Offset(x, y)
                    Dim srcBounds As New Rectangle(texnum * 64 + 63, 32, -64, 64)
                    graphicsDev.BeginScene()
                    graphicsDev.Render(destbounds, srcbounds)
                    graphicsDev.EndScene()
                    FlipToBackBuffer()
                Case "HM"
                    graphicsDev.UnprepareSource()
                    graphicsDev.BeginScene()
                    Dim oldMod As Color = graphicsDev.SetModulator(Color.Black)
                    graphicsDev.Render(heroBounds)
                    graphicsDev.SetModulator(oldMod)
                    graphicsDev.EndScene()
                    graphicsDev.PrepareSource(texture)
                    RenderHero(CType([Integer].Parse(params(0)), HeroMotion))
                    FlipToBackBuffer()
                Case "SE"
                    Instances.Current.Audio.Play(CType([Integer].Parse(params(0)), SoundEffect))
                Case "AH"
                    Dim motion As Integer = [Integer].Parse(params(0))
                    Dim x As Integer = [Integer].Parse(params(1))
                    Dim y As Integer = [Integer].Parse(params(2))
                    Dim direction As Integer = [Integer].Parse(params(3))
                    graphicsDev.BeginScene()
                    If direction = 0 Then
                        graphicsDev.Render(New Rectangle(x, y, 64, 64), New Rectangle(64 * motion, 160, 64, 64))
                    Else
                        graphicsDev.Render(New Rectangle(x, y, 64, 64), New Rectangle(64 * motion + 63, 160, -64, 64))
                    End If
                    graphicsDev.EndScene()
                    FlipToBackBuffer()
                Case "CF"
                    Dim x As Integer = [Integer].Parse(params(0))
                    Dim y As Integer = [Integer].Parse(params(1))
                    Dim w As Integer = [Integer].Parse(params(2))
                    Dim h As Integer = [Integer].Parse(params(3))
                    Dim r As Integer = [Integer].Parse(params(4))
                    Dim g As Integer = [Integer].Parse(params(5))
                    Dim b As Integer = [Integer].Parse(params(6))
                    graphicsDev.UnprepareSource()
                    graphicsDev.BeginScene()
                    Dim oldMod As Color = graphicsDev.SetModulator(Color.FromArgb(255, r, g, b))
                    graphicsDev.Render(New Rectangle(x, y, w, h))
                    graphicsDev.SetModulator(oldMod)
                    graphicsDev.EndScene()
                    graphicsDev.PrepareSource(texture)
                    FlipToBackBuffer()
                Case "LN"
                    Dim x1 As Integer = [Integer].Parse(params(0))
                    Dim y1 As Integer = [Integer].Parse(params(1))
                    Dim x2 As Integer = [Integer].Parse(params(2))
                    Dim y2 As Integer = [Integer].Parse(params(3))
                    Dim r As Integer = [Integer].Parse(params(4))
                    Dim g As Integer = [Integer].Parse(params(5))
                    Dim b As Integer = [Integer].Parse(params(6))
                    Dim gr As Graphics = workBuffer.GetGraphics()
                    Dim p As New Pen(Color.FromArgb(r, g, b))
                    gr.DrawLine(p, x1, y1, x2, y2)
                    p.Dispose()
                    workBuffer.ReleaseGraphics(gr)
                    FlipToBackBuffer()
            End Select

            FlipToScreen()

        Next aMotion

        allowRefreshing = True

    End Sub

    Private Sub FlipToBackBuffer()

        graphicsDev.Blt(backBuffer, Point.Empty, workBuffer)

    End Sub

    Private Sub FlipToScreen()

        graphicsDev.Present(backBuffer)

    End Sub

    'Ỏʂ̍XVĂꍇ True ɂȂ܂B
    Private allowRefreshing As Boolean = True

    Public Sub RefreshScreen()

        If allowRefreshing Then
            Render()
            FlipToBackBuffer()
        End If
        FlipToScreen()

    End Sub

    'ʂnÂ܂B
    Public Sub FadeOut()

        allowRefreshing = False

        Render()
        FlipToBackBuffer()
        FlipToScreen()

        graphicsDev.Fill(workBuffer, Color.Black)

        Dim destBounds As New Rectangle(0, 0, 1, 480)
        Dim srcBounds As New Rectangle(0, 0, 1, 480)

        For i As Integer = 0 To 31
            destBounds.X = i
            srcBounds.X = i
            For x As Integer = 0 To 639 Step 32
                graphicsDev.Blt(backBuffer, destBounds.Location, workBuffer, srcBounds)
                destBounds.Offset(32, 0)
                srcBounds.Offset(32, 0)
            Next x
            FlipToScreen()
            Instances.Current.Wait(30)
        Next i

        allowRefreshing = True

    End Sub

    'ʂn܂B
    Public Sub FadeIn()

        allowRefreshing = False

        Render()

        graphicsDev.Fill(backBuffer, Color.Black)
        FlipToScreen()

        Dim destBounds As New Rectangle(0, 0, 1, 480)
        Dim srcBounds As New Rectangle(0, 0, 1, 480)

        For i As Integer = 31 To 0 Step -1
            destBounds.X = i
            srcBounds.X = i
            For x As Integer = 0 To 639 Step 32
                graphicsDev.Blt(backBuffer, destBounds.Location, workBuffer, srcBounds)
                destBounds.Offset(32, 0)
                srcBounds.Offset(32, 0)
            Next x
            FlipToScreen()
            Instances.Current.Wait(30)
        Next i

        allowRefreshing = True

    End Sub

    Public Sub ShowMessage(ByVal text As String, ByVal location As PointF, ByVal emSize As Single, ByVal textColor As Color, ByVal timeout As Integer)
        ShowMessage(text, location, emSize, textColor, TimeSpan.FromMilliseconds(timeout))
    End Sub

    Public Sub ShowMessage(ByVal text As String, ByVal location As PointF, ByVal emSize As Single, ByVal textColor As Color, ByVal timeout As TimeSpan)

        allowRefreshing = False

        Render()

        Dim font As New Font("lr oSVbN", emSize, FontStyle.Bold)
        Dim brush As New SolidBrush(textColor)
        Dim g As Graphics = workBuffer.GetGraphics()
        g.DrawString(text, font, brush, location)
        workBuffer.ReleaseGraphics(g)
        brush.Dispose()
        font.Dispose()

        FlipToBackBuffer()
        FlipToScreen()

        Instances.Current.Wait(timeout)

        allowRefreshing = True

        RefreshScreen()

    End Sub

    Public Sub RenderGameOver()

        allowRefreshing = False

        graphicsDev.Fill(workBuffer, Color.White)
        Dim g As Graphics = workBuffer.GetGraphics()
        Try
            Dim font As New Font("lr o", 72.0F, FontStyle.Bold)
            Try
                g.DrawString([String].Concat("u΂", vbCrLf, "₦c"), font, Brushes.Black, 48.0F, 88.0F)
            Finally
                font.Dispose()
            End Try
        Finally
            workBuffer.ReleaseGraphics(g)
        End Try

        Try
            graphicsDev.UnprepareSource()
            graphicsDev.SetDestination(backBuffer)

            For i As Integer = 255 To 0 Step -1
                graphicsDev.Blt(backBuffer, Point.Empty, workBuffer)
                graphicsDev.BeginScene()
                graphicsDev.SetModulator(Color.FromArgb(i, 0, 0, 0))
                graphicsDev.Render(Rectangle.Empty)
                graphicsDev.SetModulator(Color.FromArgb(255, 255, 255, 255))
                graphicsDev.EndScene()
                FlipToScreen()
                Instances.Current.Wait(10)
            Next i

            Instances.Current.Wait(3000)

            For i As Integer = 0 To 255
                graphicsDev.Blt(backBuffer, Point.Empty, workBuffer)
                graphicsDev.BeginScene()
                graphicsDev.SetModulator(Color.FromArgb(i, 255, 255, 255))
                graphicsDev.Render(Rectangle.Empty)
                graphicsDev.SetModulator(Color.FromArgb(255, 255, 255, 255))
                graphicsDev.EndScene()
                FlipToScreen()
                Instances.Current.Wait(1)
            Next i

            For i As Integer = 255 To 0 Step -1
                graphicsDev.Fill(backBuffer, Color.FromArgb(i, i, i))
                FlipToScreen()
                Instances.Current.Wait(10)
            Next i

            Instances.Current.Wait(1000)

            allowRefreshing = True

        Finally
            graphicsDev.UnprepareSource()
            graphicsDev.SetDestination(workBuffer)
            graphicsDev.PrepareSource(texture)
        End Try

    End Sub

    Public Sub RenderEnding()

        allowRefreshing = False

        graphicsDev.Fill(workBuffer, Color.White)
        Dim g As Graphics = workBuffer.GetGraphics()
        Try
            Dim font As New Font("lr o", 36.0F, FontStyle.Bold)
            Try
                g.DrawString([String].Concat(Instances.Current.Hero.Name, "̒́A", vbCrLf, "ɏIcB"), font, Brushes.Black, 48.0F, 88.0F)
            Finally
                font.Dispose()
            End Try
        Finally
            workBuffer.ReleaseGraphics(g)
        End Try

        Try
            graphicsDev.UnprepareSource()
            graphicsDev.SetDestination(backBuffer)

            For i As Integer = 255 To 0 Step -1
                graphicsDev.Blt(backBuffer, Point.Empty, workBuffer)
                graphicsDev.BeginScene()
                graphicsDev.SetModulator(Color.FromArgb(i, 0, 0, 0))
                graphicsDev.Render(Rectangle.Empty)
                graphicsDev.SetModulator(Color.FromArgb(255, 255, 255, 255))
                graphicsDev.EndScene()
                FlipToScreen()
                Instances.Current.Wait(10)
            Next i

            Instances.Current.Wait(3000)

            For i As Integer = 0 To 255
                graphicsDev.Blt(backBuffer, Point.Empty, workBuffer)
                graphicsDev.BeginScene()
                graphicsDev.SetModulator(Color.FromArgb(i, 255, 255, 255))
                graphicsDev.Render(Rectangle.Empty)
                graphicsDev.SetModulator(Color.FromArgb(255, 255, 255, 255))
                graphicsDev.EndScene()
                FlipToScreen()
                Instances.Current.Wait(1)
            Next i

            For i As Integer = 255 To 0 Step -1
                graphicsDev.Fill(backBuffer, Color.FromArgb(i, i, i))
                FlipToScreen()
                Instances.Current.Wait(10)
            Next i

            Instances.Current.Wait(3000)

            Instances.Current.Audio.Play(AudioPlayer.BGMNameInEnding)

            Instances.Current.Wait(5000)

            Dim font As New Font("lr o", 24.0F, FontStyle.Bold)

            graphicsDev.Fill(workBuffer, Color.Black)
            g = workBuffer.GetGraphics()
            g.DrawString("Realistic Vision 1.0 ", font, Brushes.Pink, 48.0F, 88.0F)
            workBuffer.ReleaseGraphics(g)
            FlipToBackBuffer()
            FlipToScreen()

            Instances.Current.Wait(10000)

            graphicsDev.Fill(backBuffer, Color.Black)
            FlipToScreen()

            Instances.Current.Wait(3000)

            graphicsDev.Fill(workBuffer, Color.Black)
            g = workBuffer.GetGraphics()
            g.DrawString("Created by M.Kakumoto", font, Brushes.Pink, 48.0F, 88.0F)
            workBuffer.ReleaseGraphics(g)
            FlipToBackBuffer()
            FlipToScreen()

            Instances.Current.Wait(10000)

            graphicsDev.Fill(backBuffer, Color.Black)
            FlipToScreen()

            Instances.Current.Wait(5000)

            graphicsDev.Fill(workBuffer, Color.Black)
            g = workBuffer.GetGraphics()
            g.DrawString("Thanks for your playing!", font, Brushes.Pink, 48.0F, 88.0F)
            workBuffer.ReleaseGraphics(g)
            FlipToBackBuffer()
            FlipToScreen()

            Instances.Current.Wait(87000)

            graphicsDev.Fill(backBuffer, Color.Black)
            FlipToScreen()

            Instances.Current.Wait(3000)

            graphicsDev.Fill(workBuffer, Color.Black)
            g = workBuffer.GetGraphics()
            g.DrawString("fin...", font, Brushes.Pink, 560.0F, 440.0F)
            workBuffer.ReleaseGraphics(g)
            FlipToBackBuffer()
            FlipToScreen()

            font.Dispose()

            Instances.Current.Wait(TimeSpan.MaxValue)

            allowRefreshing = True

        Finally
            graphicsDev.UnprepareSource()
            graphicsDev.SetDestination(workBuffer)
            graphicsDev.PrepareSource(texture)
        End Try

    End Sub

    Public Sub FlashScreen()

        FlashScreen(100)

    End Sub

    Public Sub FlashScreen(ByVal timeout As Integer)

        allowRefreshing = False

        graphicsDev.Fill(backBuffer, Color.White)
        FlipToScreen()
        Instances.Current.Wait(timeout)

        allowRefreshing = True

        RefreshScreen()

    End Sub
End Class
